import numpy as np
from typing import List, Dict
from tqdm import tqdm
import networkx as nx
from .logging_utils import logger

__all__ = ['BalancedCut', 'CoalesceTree', 'make_digraph']


def make_digraph(d: np.ndarray, clust_info=None) -> nx.DiGraph:
    """
    Convert dendrogram into directed graph
    """

    g = nx.DiGraph()
    n = d.shape[0] + 1  # Dendrogram contains one less sample
    if clust_info is not None:
        if len(clust_info) != d.shape[0]+1:
            raise ValueError("ERROR: cluster information doesn't match number of leaves in dendrogram")
    else:
        clust_info = np.ones(d.shape[0]+1)*-1
    for i in tqdm(d, desc='Constructing graph from dendrogram'):
        v = i[2]  # Distance between clusters
        i = i.astype(int)
        g.add_node(n, nleaves=i[3], dist=v)
        if i[0] <= d.shape[0]:
            g.add_node(i[0], nleaves=0, dist=v, cluster=clust_info[i[0]])
        if i[1] <= d.shape[0]:
            g.add_node(i[1], nleaves=0, dist=v, cluster=clust_info[i[1]])
        g.add_edge(n, i[0])
        g.add_edge(n, i[1])
        n += 1
    if g.number_of_edges() != d.shape[0] * 2:
        logger.warning('Number of edges in directed graph not twice the dendrogram shape')
    return g


def CoalesceTree(graph: nx.DiGraph, clusters: np.ndarray) -> nx.DiGraph:

    def calc_steps_to_top(g: nx.DiGraph, c: np.ndarray):
        import pandas as pd
        s = {}
        for i in range(len(c)):
            s[i] = 0
            q = [i]
            while len(q) > 0:
                for j in g.predecessors(q.pop(0)):
                    s[i] += 1
                    q.append(j)
        return pd.Series(s).sort_values()

    def iter_predecessors(g: nx.DiGraph, v):
        q = [v]
        while len(q) > 0:
            for i in g.predecessors(q.pop(0)):
                yield i
                q.append(i)

    def aggregate_leaves(g: nx.DiGraph, v):
        q = [v]
        l = []
        while len(q) > 0:
            for i in g.successors(q.pop(0)):
                if g.nodes[i]['nleaves'] == 0:
                    l.append(i)
                else:
                    q.append(i)
        return l

    def get_holding_nodes(g: nx.DiGraph, c):
        hn = {}
        s = calc_steps_to_top(g, c)
        for i in tqdm(set(c), desc="Identifying the top node for cluster"):
            l = set(np.where(c == i)[0])
            nl = len(l)
            for j in iter_predecessors(g, s.reindex(l).idxmin()):
                if g.nodes[j]['nleaves'] >= nl:
                    l2 = aggregate_leaves(g, j)
                    if len(l.intersection(l2)) == nl:
                        hn[j] = i
                        break
        return hn

    def aggregate_predecessors(g: nx.DiGraph, v):
        p = []
        for i in iter_predecessors(g, v):
            p.append(i)
        return p

    def make_subgraph(g: nx.DiGraph, vs):

        sn = list(vs.keys())
        for i in vs:
            sn.extend(aggregate_predecessors(g, i))
        sn = list(set(sn))
        sg = nx.DiGraph(nx.subgraph(g, sn))
        for i in vs:
            sg.nodes[i]['partition_id'] = vs[i]
        return sg

    return make_subgraph(graph, get_holding_nodes(graph, clusters))


class BalancedCut:
    def __init__(self, dendrogram: np.ndarray, max_size: int, min_size: int,
                 max_distance_fc: float):
        self.nCells = dendrogram.shape[0] + 1
        self.graph = make_digraph(dendrogram)
        self.maxSize = max_size
        self.minSize = min_size
        self.maxDistFc = max_distance_fc
        self.branchpoints = self._get_branchpoints()

    def _successors(self, start: int, min_leaves: int) -> List[int]:
        """
        Get tree downstream of a node.
        """
        q = [start]
        d = []
        while len(q) > 0:
            i = q.pop(0)
            if self.graph.nodes[i]['nleaves'] > min_leaves:
                d.append(i)
                q.extend(list(self.graph.successors(i)))
        return d[1:]

    def _get_mean_dist(self, start_node: int) -> np.float:
        """
        Get mean distances in downstream tree of a node
        """
        s_nodes = self._successors(start_node, -1)
        return np.array([self.graph.nodes[x]['dist'] for x in s_nodes]).mean()

    def _are_subtrees_mergeable(self, s1: int, s2: int) -> bool:
        n1, n2 = self.graph.nodes[s1]['nleaves'], self.graph.nodes[s2]['nleaves']
        if n1 > self.minSize and n2 > self.minSize:
            d1, d2 = self.graph.nodes[s1]['dist'], self.graph.nodes[s2]['dist']
            if d1/d2 > self.maxDistFc or d2/d1 > self.maxDistFc:
                logger.debug(f"Will not merge {s1} and {s2} because of high distance")
                return False
            else:
                md1, md2 = self._get_mean_dist(s1), self._get_mean_dist(s2)
                if md1/md2 > self.maxDistFc or md2/md1 > self.maxDistFc:
                    logger.debug(f"Will not merge {s1} and {s2} because of high distance of successors")
                    return False
        return True

    def _get_branchpoints(self) -> \
            Dict[int, List[int]]:
        """
        Aggregate leaves bottom up until target size is reached.
        """
        n_leaves = int((self.graph.number_of_nodes() + 1) / 2)
        leaves = {x: None for x in range(n_leaves)}
        bps = {}
        pbar = tqdm(total=len(leaves), desc='Identifying nodes to split')
        while len(leaves) > 0:
            leaf, _ = leaves.popitem()
            pbar.update(1)
            logger.debug(f"FRESH STEP: Leaf {leaf} plucked as base leaf")
            cur = leaf
            while True:
                temp = next(self.graph.predecessors(cur))
                if temp in bps:
                    logger.debug(f"Will not climb to {temp} as already a branchpoint")
                    break
                if self.graph.nodes[temp]['nleaves'] > self.maxSize:
                    logger.debug(f"Will not climb to {temp} because too many leaves exist")
                    break
                s1, s2 = list(self.graph.successors(temp))
                if self._are_subtrees_mergeable(s1, s2) is False:
                    break
                cur = temp
            logger.debug(f"Aggregating from branch {cur} for leaf {leaf}")
            bps[cur] = [leaf]
            s = [cur]
            while len(s) > 0:
                i = s.pop()
                if i in leaves:
                    bps[cur].append(i)
                    leaves.pop(i)
                    logger.debug(f"Leaf {i} plucked in aggregation step")
                    pbar.update(1)
                elif i in bps and i != cur:
                    logger.debug(f"Skipping branch {i} because its already taken")
                elif self.graph.nodes[i]['nleaves'] >= self.maxSize and i != cur:
                    logger.debug(f"Skipping branch {i} to prevent greedy behaviour")
                else:
                    s.extend(list(self.graph.successors(i)))
        pbar.close()
        return bps

    def _valid_names_in_branchpoints(self) -> None:
        leaves = []
        for i in self.branchpoints:
            leaves.extend(self.branchpoints[i])
        n_leaves = len(leaves)
        if n_leaves != self.nCells:
            raise ValueError("ERROR: Not all leaves present in branchpoints. This bug must be reported")
        minl = min(leaves)
        if minl != 0:
            raise ValueError(f"ERROR: minimum leaf label is {minl} rather than 0")
        maxl = max(leaves)
        if n_leaves != maxl+1:
            raise ValueError(f"ERROR: maximum leaf label is {maxl} while total estimated leaves are {n_leaves}")
        return None

    def get_clusters(self) -> np.ndarray:
        """
        Make cluster label array from `get_branchpoints` output
        """
        self._valid_names_in_branchpoints()
        c = np.zeros(self.nCells).astype(int)
        for n, i in enumerate(self.branchpoints, start=1):
            c[self.branchpoints[i]] = n
        if (c == 0).sum() > 0:
            logger.warning(f"{(c == 0).sum()} samples were not assigned a cluster")
            c[c == 0] = -1
        return c

    # def get_sibling(g, node):
    #     sibs = g.successors(next(g.predecessors(node)))
    #     return [x for x in sibs if x != node][0]

    # def top_rsquared(d, cutoff):
    #     df = pd.DataFrame([list(range(len(d))), d.values], index=['pos', 'val']).T
    #     f ='pos~val'
    #     f ='val~pos'
    #     rs = []
    #     for i in range(1, len(d)-2):
    #         rs.append(ols(formula=f, data=df[:-i]).fit().rsquared)
    #     rs = np.array(rs)
    #     pos = np.where((rs > cutoff) == True)[0][0]
    #     return 1+pos

    # def merge_branches(g, bps, nodes):
    #     bps = dict(bps)
    #     for i in nodes:
    #         if i not in bps:
    #             logger.debug(f"Ignoring {i} because not a branchpoint anymore")
    #         sib = get_sibling(g, i)
    #         if sib in bps:
    #             if sib in nodes:
    #                 logger.debug(f"Merging two sibling candidates {i} and {sib}")
    #                 bps[next(g.predecessors(i))] = bps[i] + bps[sib]
    #                 del bps[i]
    #                 del bps[sib]
    #             else:
    #                 logger.debug(f"Branchpoint {sib} updated with values from {i}")
    #                 bps[sib] = bps[i] + bps[sib]
    #                 del bps[i]
    #         else:
    #             logger.debug(f"Not merging {i} because sibling not a branchpoint")
    #     return bps

    # def test(self):
        # n = 30
        # np.random.seed(154)
        # randz = ward(gamma.rvs(0.3, size=n * 4).reshape((n, 4)))
        #
        # graph = z_to_g(randz)
        # branchpoints = get_branchpoints(graph, max_leaf=10, min_leaves=2, fc=1.5)
        # clusters = bps_to_clusts(branchpoints)
        # dend = dendrogram(randz)
        # clusters[dend['leaves']]
        #
        # res = [9, 5, 5, 1, 1, 1, 1, 7, 7, 7, 3, 3, 3, 3, 3, 4, 4, 4, 2, 2, 2, 8,
        #        8, 8, 6, 6, 6, 6, 6, 6]



"""
Utility functions for running the KNN algorithm.
"""
import numpy as np
from .writers import create_zarr_dataset
from .ann import AnnStream
from tqdm import tqdm
from .logging_utils import logger
import pandas as pd

__all__ = ['self_query_knn', 'smoothen_dists', 'export_knn_to_mtx']


def self_query_knn(ann_obj: AnnStream, store, chunk_size: int, nthreads: int) -> None:
    """
    Constructs KNN graph.

    Args:
        ann_obj ():
        store ():
        chunk_size ():
        nthreads (): Number of threads to use.

    Returns:
        None
    """
    from threadpoolctl import threadpool_limits

    n_cells, n_neighbors = ann_obj.nCells, ann_obj.k
    z_knn = create_zarr_dataset(store, 'indices', (chunk_size,), 'u8',
                                (n_cells, n_neighbors))
    z_dist = create_zarr_dataset(store, 'distances', (chunk_size,), 'f8',
                                 (n_cells, n_neighbors))
    nsample_start = 0
    tnm = 0  # Number of missed recall
    with threadpool_limits(limits=nthreads):
        for i in ann_obj.iter_blocks(msg='Saving KNN graph'):
            nsample_end = nsample_start + i.shape[0]
            ki, kv, nm = ann_obj.transform_ann(ann_obj.reducer(i), k=n_neighbors,
                                               self_indices=np.arange(nsample_start, nsample_end))
            z_knn[nsample_start:nsample_end, :] = ki
            z_dist[nsample_start:nsample_end, :] = kv
            nsample_start = nsample_end
            tnm += nm
    recall = ann_obj.data.shape[0] - tnm
    recall = 100 * recall / ann_obj.data.shape[0]
    recall = "%.2f" % recall
    logger.info(f"ANN recall: {recall}%")
    return None


def _is_umap_version_new():
    import umap
    from packaging import version

    if version.parse(umap.__version__) >= version.parse('0.5.0'):
        return True
    else:
        return False


def smoothen_dists(store, z_idx, z_dist, lc: float, bw: float, chunk_size: int = 100000):
    """
    Smoothens KNN distances.

    Args:
        store ():
        z_idx ():
        z_dist ():
        lc ():
        bw ():
        chunk_size ():

    Returns:
        None

    """
    from umap.umap_ import smooth_knn_dist, compute_membership_strengths

    umap_is_latest = _is_umap_version_new()

    n_cells, n_neighbors = z_idx.shape
    zge = create_zarr_dataset(store, f'edges', (chunk_size,), ('u8', 'u8'),
                              (n_cells * n_neighbors, 2))
    zgw = create_zarr_dataset(store, f'weights', (chunk_size,), 'f8',
                              (n_cells * n_neighbors,))
    last_row = 0
    val_counts = 0
    step = int(chunk_size / n_neighbors)
    for i in tqdm(range(0, n_cells, step), desc='Smoothening KNN distances'):
        if i + step > n_cells:
            ki, kv = z_idx[i:n_cells, :], z_dist[i:n_cells, :]
        else:
            ki, kv = z_idx[i:i+step, :], z_dist[i:i+step, :]
        kv = kv.astype(np.float32, order='C')
        sigmas, rhos = smooth_knn_dist(kv, k=n_neighbors,
                                       local_connectivity=lc, bandwidth=bw)
        if umap_is_latest:
            rows, cols, vals, _ = compute_membership_strengths(ki, kv, sigmas, rhos)
        else:
            rows, cols, vals = compute_membership_strengths(ki, kv, sigmas, rhos)
        rows = rows + last_row
        start = val_counts
        end = val_counts + len(rows)
        last_row = rows[-1] + 1
        val_counts += len(rows)
        zge[start:end, 0] = rows
        zge[start:end, 1] = cols
        zgw[start:end] = vals
    # Fixing edges with 0 weights
    w = zgw[:]
    idx = w == 0
    minv = w[~idx].min()
    w[idx] = minv
    zgw[:] = w
    return None


def export_knn_to_mtx(mtx: str, csr_graph, batch_size: int = 1000) -> None:
    """
    Exports KNN matrix in Matrix Market format.

    Args:
        mtx:
        csr_graph:
        batch_size:

    Returns:
        None

    """
    n_cells = csr_graph.shape[0]
    with open(mtx, 'w') as h:
        h.write("%%MatrixMarket matrix coordinate real general\n% Generated by Scarf\n")
        h.write(f"{n_cells} {n_cells} {csr_graph.nnz}\n")
        s = 0
        for e in tqdm(range(batch_size, n_cells + batch_size, batch_size),
                      desc='Saving KNN matrix in MTX format'):
            if e > n_cells:
                e = n_cells
            sg = csr_graph[s:e].tocoo()
            df = pd.DataFrame({'row': sg.row + s + 1, 'col': sg.col + 1, 'd': sg.data})
            df.to_csv(h, sep=' ', header=False, index=False, mode='a')
            s = e
        if s != n_cells:
            raise ValueError("ERROR: Internal loop count error in export_knn_to_mtx. Please report this bug")
    return None



"""
Contains the code for plotting in Scarf.
"""
from re import A
import matplotlib.pyplot as plt
import matplotlib as mpl
import seaborn as sns
import numpy as np
import pandas as pd
from typing import Tuple, Optional
from cmocean import cm
from .logging_utils import logger

plt.style.use('fivethirtyeight')
plt.rcParams['svg.fonttype'] = 'none'


def clean_axis(ax, ts=11, ga=0.4):
    """
    Cleans a given matplotlib axis.
    """
    ax.xaxis.set_tick_params(labelsize=ts)
    ax.yaxis.set_tick_params(labelsize=ts)
    for i in ['top', 'bottom', 'left', 'right']:
        ax.spines[i].set_visible(False)
    ax.grid(which='major', linestyle='--', alpha=ga)
    ax.figure.patch.set_alpha(0)
    ax.patch.set_alpha(0)
    return True


def plot_graph_qc(g):
    # TODO: add docstring description. Is this for qc of a graph, or for plotting a qc plot of a graph?
    _, axis = plt.subplots(1, 2, figsize=(12, 4))
    ax = axis[0]
    x = np.array((g != 0).sum(axis=0))[0]
    y = pd.Series(x).value_counts().sort_index()
    ax.bar(y.index, y.values, width=0.5)
    xlim = np.percentile(x, 99.5) + 5
    ax.set_xlim((0, xlim))
    ax.set_xlabel('Node degree')
    ax.set_ylabel('Frequency')
    ax.text(xlim, y.values.max(), f"plot is clipped (max degree: {y.index.max()})",
            ha='right', fontsize=9)
    clean_axis(ax)
    ax = axis[1]
    ax.hist(g.data, bins=30)
    ax.set_xlabel('Edge weight')
    ax.set_ylabel('Frequency')
    clean_axis(ax)
    plt.tight_layout()
    plt.show()


def plot_qc(data: pd.DataFrame, color: str = 'steelblue', cmap: str = 'tab20',
            fig_size: tuple = None, label_size: float = 10.0, title_size: float = 10,
            sup_title: str = None, sup_title_size: float = 12, scatter_size: float = 1.0,
            max_points: int = 10000, show_on_single_row: bool = True):
    # TODO: add docstring description. Is this for qc of a plot, or for plotting a qc plot?
    n_plots = data.shape[1] - 1
    n_groups = data['groups'].nunique()
    if n_groups > 5 and show_on_single_row is True:
        logger.info(f"Too many groups in the plot. If you think that plot is too wide then consider turning "
                    f"`show_on_single_row` parameter to False")
    if show_on_single_row is True:
        n_rows = 1
        n_cols = n_plots
    else:
        n_rows = n_plots
        n_cols = 1
    if fig_size is None:
        fig_width = min(15, n_groups+(2*n_cols))
        fig_height = 1+2.5*n_rows
        fig_size = (fig_width, fig_height)
    fig = plt.figure(figsize=fig_size)
    grouped = data.groupby('groups')
    for i in range(n_plots):
        if data.columns[i] == 'groups':
            continue
        vals = {'g': [], 'v': []}
        for j in sorted(data['groups'].unique()):
            val = grouped.get_group(j)[data.columns[i]].values
            vals['g'].extend([j for _ in range(len(val))])
            vals['v'].extend(list(val))
        vals = pd.DataFrame(vals)
        ax = fig.add_subplot(n_rows, n_cols, i+1)
        if n_groups == 1:
            sns.violinplot(y='v', x='g', data=vals, linewidth=1, orient='v', alpha=0.6,
                           inner=None, cut=0, color=color)
        else:
            sns.violinplot(y='v', x='g', data=vals, linewidth=1, orient='v', alpha=0.6,
                           inner=None, cut=0, palette=cmap)
        if len(vals) > max_points:
            sns.stripplot(x='g', y='v', data=vals.sample(n=max_points), jitter=0.4, ax=ax, orient='v',
                          s=scatter_size, color='k', alpha=0.4)
        else:
            sns.stripplot(x='g', y='v', data=vals, jitter=0.4, ax=ax, orient='v',
                          s=scatter_size, color='k', alpha=0.4)
        ax.set_ylabel(data.columns[i], fontsize=label_size)
        ax.set_xlabel('')
        if n_groups == 1:
            ax.set_xticklabels([])
        if data['groups'].nunique() == 1:
            ax.set_title('Median: %.1f' % (int(np.median(vals['v']))), fontsize=title_size)
        clean_axis(ax)
    fig.suptitle(sup_title, fontsize=sup_title_size)
    plt.tight_layout()
    plt.show()


def plot_mean_var(nzm: np.ndarray, fv: np.ndarray, n_cells: np.ndarray, hvg: np.ndarray,
                  ax_label_fs: float = 12, fig_size: Tuple[float, float] = (4.5, 4.0),
                  ss: Tuple[float, float] = (3, 30), cmaps: Tuple[str, str] = ('winter', 'magma_r')):
    """
    Shows a mean-variance plot.
    """
    _, ax = plt.subplots(1, 1, figsize=fig_size)
    nzm = np.log2(nzm)
    fv = np.log2(fv)
    ax.scatter(nzm[~hvg], fv[~hvg], alpha=0.6, c=n_cells[~hvg], cmap=cmaps[0], s=ss[0])
    ax.scatter(nzm[hvg], fv[hvg], alpha=0.8, c=n_cells[hvg], cmap=cmaps[1], s=ss[1], edgecolor='k', lw=0.5)
    ax.set_xlabel('Log mean non-zero expression', fontsize=ax_label_fs)
    ax.set_ylabel('Log corrected variance', fontsize=ax_label_fs)
    clean_axis(ax)
    plt.tight_layout()
    plt.show()


def plot_elbow(var_exp, figsize: Tuple[float, float] = (None, 2)):
    from kneed import KneeLocator

    x = range(len(var_exp))
    kneedle = KneeLocator(x, var_exp, S=1.0, curve="convex", direction="decreasing")
    if figsize[0] is None:
        figsize = (0.25 * len(var_exp), figsize[1])
    fig, ax = plt.subplots(1, 1, figsize=figsize)
    ax.plot(x, var_exp, lw=1)
    ax.set_xticks(x)
    ax.axvline(kneedle.elbow, lw=1, c='r', label='Elbow')
    ax.set_ylabel("% Variance explained", fontsize=9)
    ax.set_xlabel("Principal components", fontsize=9)
    clean_axis(ax, ts=8)
    ax.legend(frameon=False, fontsize=9)
    plt.tight_layout()
    plt.show()


def plot_heatmap(cdf, fontsize: float = 10, width_factor: float = 0.03, height_factor: float = 0.02,
                 cmap=cm.matter_r, savename: str = None, save_dpi: int = 300, figsize=None):
    """
    Shows a heatmap plot.
    """
    if figsize is None:
        figsize = (cdf.shape[1]*fontsize*width_factor, fontsize*cdf.shape[0]*height_factor)
    cgx = sns.clustermap(cdf, yticklabels=cdf.index, xticklabels=cdf.columns, method='ward',
                         figsize=figsize, cmap=cmap, rasterized=True)
    cgx.ax_heatmap.set_yticklabels(cdf.index[cgx.dendrogram_row.reordered_ind], fontsize=fontsize)
    cgx.ax_heatmap.set_xticklabels(cdf.columns[cgx.dendrogram_col.reordered_ind], fontsize=fontsize)
    cgx.ax_heatmap.figure.patch.set_alpha(0)
    cgx.ax_heatmap.patch.set_alpha(0)
    if savename:
        plt.savefig(savename, dpi=save_dpi)
    plt.show()
    return None


def _scatter_fix_type(v: pd.Series, ints_as_cats: bool) -> pd.Series:
    vt = v.dtype
    if v.nunique() == 1:
        return pd.Series(np.ones(len(v)), index=v.index).astype(np.float_)
    if vt in [np.bool_]:
        # converting first to int to handle bool
        return v.astype(np.int_).astype('category')
    if vt in [str, object] or vt.name == 'category':
        return v.astype('category')
    elif np.issubdtype(vt.type, np.integer) and ints_as_cats:
        if v.nunique() > 100:
            logger.warning("Too many categories. set force_ints_as_cats to false")
        return v.astype(np.int_).astype('category')
    else:
        return v.astype(np.float_)


def _scatter_fix_mask(v: pd.Series, mask_vals: list, mask_name: str) -> pd.Series:
    if mask_vals is None:
        mask_vals = []
    mask_vals += [np.NaN]
    iscat = False
    if v.dtype.name == 'category':
        iscat = True
        v = v.astype(object)
    # There is a bug in pandas which causes failure above 1M rows
    # v[v.isin(mask_vals)] = mask_name
    v[np.isin(v, mask_vals)] = mask_name
    if iscat:
        v = v.astype('category')
    return v


def _scatter_make_colors(v: pd.Series, cmap, color_key: Optional[dict], mask_color: str, mask_name: str):
    from matplotlib.cm import get_cmap

    if v.dtype.name != 'category':
        if cmap is None:
            return cm.deep, None
        else:
            return get_cmap(cmap), None
    else:
        if cmap is None:
            cmap = 'tab20'
    na_idx = v == mask_name
    uv = v[~na_idx].unique()
    if color_key is not None:
        for i in uv:
            if i not in color_key:
                raise KeyError(f"ERROR: key {i} missing in `color_key`")
        if na_idx.sum() > 0:
            if mask_name not in color_key:
                color_key[mask_name] = mpl.colors.to_hex(mask_color)
        return None, color_key
    else:
        pal = sns.color_palette(cmap, n_colors=len(uv)).as_hex()
        color_key = dict(zip(sorted(uv), pal))
        if na_idx.sum() > 0:
            color_key[mask_name] = mpl.colors.to_hex(mask_color)
        return None, color_key


def _scatter_cleanup(ax, sw: float, sc: str, ds: tuple) -> None:
    for i in ['bottom', 'left', 'top', 'right']:
        spine = ax.spines[i]
        if i in ds:
            spine.set_visible(True)
            spine.set_linewidth(sw)
            spine.set_edgecolor(sc)
        else:
            spine.set_visible(False)
    ax.figure.patch.set_alpha(0)
    ax.patch.set_alpha(0)
    ax.set_aspect('auto')
    return None


def _scatter_label_axis(df, ax, fs: float, fo: float):
    x, y = df.columns[:2]
    ax.set_xlabel(x, fontsize=fs)
    ax.set_ylabel(y, fontsize=fs)
    vmin, vmax = df[x].min(), df[x].max()
    ax.set_xlim((vmin - abs(vmin * fo), vmax + abs(vmax * fo)))
    vmin, vmax = df[y].min(), df[y].max()
    ax.set_ylim((vmin - abs(vmin * fo), vmax + abs(vmax * fo)))
    ax.set_xticks([])
    ax.set_yticks([])
    return None


def _scatter_legends(df, ax, fig, cmap, ck, ondata: bool, onside: bool, fontsize: float,
                     n_per_col: int, scale: float, ls: float, cs: float) -> None:
        from matplotlib.colors import Normalize
        from matplotlib.colorbar import ColorbarBase
        from mpl_toolkits.axes_grid1.axes_divider import make_axes_locatable


        x, y, vc = df.columns[:3]
        v = df[vc]
        if v.nunique() <= 1:
            return None
        if v.dtype.name == 'category':
            ax.title.set_text(vc)
            centers = df[[x, y, vc]].groupby(vc).median().T
            for i in centers:
                if ondata:
                    ax.text(centers[i][x], centers[i][y], i, fontsize=fontsize,
                            ha='center', va='center')
                if onside:
                    ax.scatter([float(centers[i][x])], [float(centers[i][y])],
                               c=ck[i], label=i, alpha=1, s=0.01)
            if onside:
                n_cols = v.nunique() // n_per_col
                if v.nunique() % n_per_col > 0:
                    n_cols += 1
                ax.legend(ncol=n_cols, loc=(1, 0), frameon=False, fontsize=fontsize,
                          markerscale=scale, labelspacing=ls, columnspacing=cs)
        else:
            if fig is not None:
                ax_divider = make_axes_locatable(ax)
                cax = ax_divider.append_axes("top", size="7%", pad="5%")
                norm = Normalize(vmin=v.min(), vmax=v.max())
                cb = ColorbarBase(cax, cmap=cmap, norm=norm, orientation='horizontal')
                cb.set_label(vc, fontsize=fontsize)
                cb.ax.xaxis.set_label_position('top')
                fig.set_constrained_layout_pads(wspace=0.1, hspace=0.05)
            else:
                logger.warning("Not plotting the colorbar because fig object was not passed")
        return None


def _make_grid(width, height, w_pad, h_pad, n_panels, n_columns):
    n_columns = np.minimum(n_panels, n_columns)
    n_rows = np.ceil(n_panels / n_columns).astype(int)
    if (w_pad is None and h_pad is None):
        constrained = True
    else:
        constrained = False
    fig, axes = plt.subplots(n_rows, n_columns, figsize=(width * n_columns, height * n_rows), squeeze=False, constrained_layout=constrained)
    diff = (n_rows * n_columns) - n_panels
    while diff > 0:
        fig.delaxes(axes[n_rows - 1, n_columns - diff])
        diff -= 1

    if not constrained:
        plt.tight_layout(w_pad=w_pad, h_pad=h_pad)

    return fig, axes


def plot_scatter(dfs, in_ax=None, fig=None, width: float = 6, height: float = 6,
                      default_color: str = 'steelblue', color_map=None, color_key: dict = None,
                      mask_values: list = None, mask_name: str = 'NA', mask_color: str = 'k', 
                      point_size: float = 10, ax_label_size: float = 12, frame_offset: float = 0.05,
                      spine_width: float = 0.5, spine_color: str = 'k', displayed_sides: tuple = ('bottom', 'left'),
                      legend_ondata: bool = True, legend_onside: bool = True,
                      legend_size: float = 12, legends_per_col: int = 20,
                      marker_scale: float = 70, lspacing: float = 0.1, cspacing: float = 1,
                      savename: str = None, dpi: int = 300, force_ints_as_cats: bool = True,
                      n_columns: int = 4, w_pad: float = None, h_pad: float = None, scatter_kwargs: dict = None):
    """
    Shows scatter plots. If more then one dataframe is provided it will place the scatterplots in a grid. 
    """
    from matplotlib.colors import to_hex

    def _handle_scatter_kwargs(sk):
        if sk is None:
            sk = {}
        if 'c' in sk:
            logger.warning('scatter_kwarg value `c` will be ignored')
            del sk['c']
        if 's' in sk:
            logger.warning('scatter_kwarg value `s` will be ignored')
            del sk['s']
        if 'lw' not in sk:
            sk['lw'] = 0.1
        if 'edgecolors' not in sk:
            sk['edgecolors'] = 'k'
        return sk

    if len(dfs) > 1:
        fig, axs = _make_grid(width, height, w_pad, h_pad, len(dfs), n_columns)
    else:
        fig, axs = plt.subplots(1, 1, figsize=(width, height), squeeze=False)

    for i, df in enumerate(dfs):
        dim1, dim2, vc = df.columns[:3]
        v = _scatter_fix_mask(df[vc].copy(), mask_values, mask_name)
        v = _scatter_fix_type(v, force_ints_as_cats)
        df[vc] = v
        col_map, col_key = _scatter_make_colors(v, color_map, color_key,
                                                    mask_color, mask_name)
        if v.dtype.name == 'category':
            df['c'] = [col_key[x] for x in v]
        else:
            if v.nunique() == 1:
                df['c'] = [default_color for _ in v]
            else:
                v = v.copy().fillna(0)
                # FIXME: Why have the following line?
                pal = col_map
                mmv = (v - v.min()) / (v.max() - v.min())
                df['c'] = [to_hex(pal(x)) for x in mmv]
        if 's' not in df:
            df['s'] = [point_size for _ in df.index]
        scatter_kwargs = _handle_scatter_kwargs(sk=scatter_kwargs)
        ax = axs[int(i / n_columns), i % n_columns]
        ax.scatter(df[dim1].values, df[dim2].values, c=df['c'].values, s=df['s'].values,
                rasterized=True, **scatter_kwargs)
        _scatter_label_axis(df, ax, ax_label_size, frame_offset)
        _scatter_cleanup(ax, spine_width, spine_color, displayed_sides)
        _scatter_legends(df, ax, fig, col_map, col_key, legend_ondata, legend_onside,
                        legend_size, legends_per_col, marker_scale, lspacing, cspacing)

    if in_ax is None:
        if savename:
            plt.savefig(savename, dpi=dpi, bbox_inches='tight')
        plt.show()
    else:
        return axs


def shade_scatter(dfs, in_ax=None, figsize: float = 6, pixels: int = 1000, sampling: float = 0.1,
                  spread_px: int = 1, spread_threshold: float = 0.2, min_alpha: int = 10, 
                  color_map=None, color_key: dict = None,
                  mask_values: list = None, mask_name: str = 'NA', mask_color: str = 'k', 
                  ax_label_size: float = 12, frame_offset: float = 0.05,
                  spine_width: float = 0.5, spine_color: str = 'k', displayed_sides: tuple = ('bottom', 'left'),
                  legend_ondata: bool = True, legend_onside: bool = True,
                  legend_size: float = 12, legends_per_col: int = 20,
                  marker_scale: float = 70, lspacing: float = 0.1, cspacing: float = 1,
                  savename: str = None, dpi: int = 300, force_ints_as_cats: bool = True,
                  n_columns: int = 4, w_pad: float = None, h_pad: float = None):
    """
    Shows shaded scatter plots. If more then one dataframe is provided it will place the scatterplots in a grid. 
    """
    import datashader as dsh
    from IPython.display import display
    from datashader.mpl_ext import dsshow
    import datashader.transfer_functions as tf
    from functools import partial
    
    if len(dfs) > 1:
        fig, axs = _make_grid(figsize, figsize, w_pad, h_pad, len(dfs), n_columns)
    else:
        fig, axs = plt.subplots(1, 1, figsize=(figsize, figsize), squeeze=False)

    for i, df in enumerate(dfs):
        dim1, dim2, vc = df.columns[:3]
        v = _scatter_fix_mask(df[vc].copy(), mask_values, mask_name)
        v = _scatter_fix_type(v, force_ints_as_cats)
        df[vc] = v
        col_map, col_key = _scatter_make_colors(v, color_map, color_key,
                                                    mask_color, mask_name)
        if v.dtype.name == 'category':
            agg = dsh.count_cat(vc)
        else:
            if v.nunique() == 1:
                agg = dsh.count(vc)
            else:
                agg = dsh.mean(vc)
    
        ax = axs[int(i / n_columns), i % n_columns]
        artist = dsshow(df, dsh.Point(dim1, dim2), aggregator=agg, norm='eq_hist', 
                        color_key=col_key, cmap=col_map, alpha_range=(min_alpha, 255),
                        shade_hook=partial(tf.dynspread, threshold=spread_threshold, max_px=spread_px), 
                        plot_height=pixels, plot_width=pixels, aspect='equal', width_scale=1, height_scale=1,
                        ax=ax)

        _scatter_label_axis(df, ax, ax_label_size, frame_offset) 
        _scatter_cleanup(ax, spine_width, spine_color, displayed_sides)
        _scatter_legends(df, ax, fig, col_map, col_key, legend_ondata, legend_onside,
                        legend_size, legends_per_col, marker_scale, lspacing, cspacing)
    
    if savename:
        plt.savefig(savename, dpi=dpi, bbox_inches='tight')
    if in_ax is not None:
        return axs

def _draw_pie(ax, dist, colors, xpos, ypos, size):
    # https://stackoverflow.com/questions/56337732/how-to-plot-scatter-pie-chart-using-matplotlib
    cumsum = np.cumsum(dist)
    cumsum = cumsum / cumsum[-1]
    pie = [0] + cumsum.tolist()
    for r1, r2, c in zip(pie[:-1], pie[1:], colors):
        angles = np.linspace(2 * np.pi * r1, 2 * np.pi * r2)
        x = [0] + np.cos(angles).tolist()
        y = [0] + np.sin(angles).tolist()
        xy = np.column_stack([x, y])
        ax.scatter([xpos], [ypos], marker=xy, s=size, c=c)


def plot_cluster_hierarchy(sg, clusts, color_values=None, force_ints_as_cats: bool = True,
                           width: float = 2, lvr_factor: float = 0.5, vert_gap: float = 0.2,
                           min_node_size: float = 10, node_size_multiplier: float = 1e4, node_power: float = 1,
                           root_size: float = 100, non_leaf_size: float = 10,
                           show_labels: bool = False, fontsize=10,
                           root_color: str = '#C0C0C0', non_leaf_color: str = 'k',
                           cmap: str = None, color_key: bool = None, edgecolors: str = 'k',
                           edgewidth: float = 1, alpha: float = 0.7, figsize=(5, 5), ax=None, show_fig: bool = True,
                           savename: str = None, save_dpi=300):
    """
    Shows a plot showing cluster hierarchy.

    Returns:
        If requested (with parameter `show_fig`) a matplotlib Axes object containing the plot
        (which is the modified `ax` parameter if given).
    """
    import networkx as nx
    import EoN
    import math
    from matplotlib.colors import to_hex

    if color_values is None:
        color_values = pd.Series(clusts)
        using_clust_for_colors = True
    else:
        color_values = pd.Series(color_values)
        using_clust_for_colors = False
    color_values = _scatter_fix_type(color_values, force_ints_as_cats)
    cmap, color_key = _scatter_make_colors(color_values, cmap, color_key,
                                           'k', 'longdummyvaluesofh3489hfpiqehdcbla')
    pos = EoN.hierarchy_pos(sg, width=width * math.pi, leaf_vs_root_factor=lvr_factor, vert_gap=vert_gap)
    new_pos = {u: (r * math.cos(theta), r * math.sin(theta)) for u, (theta, r) in pos.items()}

    if color_key is None:
        cluster_values = pd.DataFrame({'clusters': clusts, 'v': color_values}).groupby('clusters').mean()['v']
        mmv: pd.Series = (cluster_values - cluster_values.min()) / (cluster_values.max() - cluster_values.min())
        color_key = {k: to_hex(cmap(v)) for k, v in mmv.to_dict().items()}
    else:
        cluster_values = None

    cs = pd.Series(clusts).value_counts()
    cs = (node_size_multiplier * ((cs / cs.sum()) ** node_power)).to_dict()
    nc, ns = [], []

    for i in sg.nodes():
        if 'partition_id' in sg.nodes[i]:
            clust_id = sg.nodes[i]['partition_id']
            if cluster_values is not None or using_clust_for_colors:
                nc.append(color_key[clust_id])
                ns.append(max(cs[clust_id], min_node_size))
            else:
                nc.append('white')
                ns.append(0)
        else:
            if sg.nodes[i]['nleaves'] == len(clusts):
                nc.append(root_color)
                ns.append(root_size)
            else:
                nc.append(non_leaf_color)
                ns.append(non_leaf_size)
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=figsize)
    nx.draw(sg, pos=new_pos, node_size=ns, node_color=nc, ax=ax, edgecolors=edgecolors, alpha=alpha,
            linewidths=edgewidth)

    if cluster_values is None and using_clust_for_colors is False:
        for i in sg.nodes():
            if 'partition_id' in sg.nodes[i]:
                clust_id = sg.nodes[i]['partition_id']
                idx = clusts == clust_id
                counts = color_values[idx].value_counts()
                _draw_pie(ax, counts.values, [color_key[x] for x in counts.index],
                          new_pos[i][0], new_pos[i][1], max(cs[clust_id], min_node_size))

    if show_labels:
        for i in sg.nodes():
            if 'partition_id' in sg.nodes[i]:
                clust_id = sg.nodes[i]['partition_id']
                ax.text(new_pos[i][0], new_pos[i][1], clust_id, fontsize=fontsize, ha='center', va='center')
    if savename:
        plt.savefig(savename, dpi=save_dpi)
    if show_fig:
        plt.show()
    else:
        return ax



"""
Utility methods.

- Methods:
    - clean_array: returns input array with nan and infinite values removed
    - controlled_compute: performs computation with Dask
    - rescale_array: performs edge trimming on values of the input vector
    - show_progress: performs computation with Dask and shows progress bar
    - system_call: executes a command in the underlying operative system
"""

import numpy as np
from .logging_utils import logger

__all__ = ['system_call', 'rescale_array', 'clean_array', 'show_progress', 'controlled_compute']


def rescale_array(a: np.ndarray, frac: float = 0.9) -> np.ndarray:
    """
    Performs edge trimming on values of the input vector.

    Performs edge trimming on values of the input vector and constraints them between within frac and 1-frac density of
    normal distribution created with the sample mean and std. dev. of a.

    Args:
        a: numeric vector
        frac: Value between 0 and 1.

    Return:
        The input array, edge trimmed and constrained.
    """
    from scipy.stats import norm

    loc = (np.median(a) + np.median(a[::-1])) / 2
    dist = norm(loc, np.std(a))
    minv, maxv = dist.ppf(1 - frac), dist.ppf(frac)
    a[a < minv] = minv
    a[a > maxv] = maxv
    return a


def clean_array(x, fill_val: int = 0):
    """
    Returns input array with nan and infinite values removed.

    Args:
        x (np.ndarray): input array
        fill_val: value to fill zero values with (default: 0)
    """
    x = np.nan_to_num(x, copy=True)
    x[(x == np.Inf) | (x == -np.Inf)] = 0
    x[x == 0] = fill_val
    return x


def controlled_compute(arr, nthreads):
    """
    Performs computation with Dask.

    Args:
        arr:
        nthreads: number of threads to use for computation

    Returns:
        Result of computation.
    """
    from multiprocessing.pool import ThreadPool
    import dask

    with dask.config.set(schedular='threads', pool=ThreadPool(nthreads)):
        res = arr.compute()
    return res


def show_progress(arr, msg: str = None, nthreads: int = 1):
    """
    Performs computation with Dask and shows progress bar.

    Args:
        arr:
        msg: message to log, default None
        nthreads: number of threads to use for computation, default 1

    Returns:
        Result of computation.
    """
    from dask.diagnostics import ProgressBar

    if msg is not None:
        logger.info(msg)
    pbar = ProgressBar()
    pbar.register()
    res = controlled_compute(arr, nthreads)
    pbar.unregister()
    return res


def system_call(command):
    """Executes a command in the underlying operative system."""
    import subprocess
    import shlex

    process = subprocess.Popen(shlex.split(command), stdout=subprocess.PIPE)
    while True:
        output = process.stdout.readline()
        if process.poll() is not None:
            break
        if output:
            logger.info(output.strip())
    process.poll()
    return None



"""
- Classes:
    - Assay: A generic Assay class that contains methods to calculate feature level statistics.
    - RNAassay: This assay is designed for feature selection and normalization of scRNA-Seq data.
    - ATACassay:
    - ADTassay:
"""
# TODO: add description to docstring

import numpy as np
import dask.array as daskarr
import zarr
from .metadata import MetaData
from .utils import show_progress, controlled_compute
from scipy.sparse import csr_matrix, vstack
from .logging_utils import logger
from typing import Tuple, List
import pandas as pd

__all__ = ['Assay', 'RNAassay', 'ATACassay', 'ADTassay']


def norm_dummy(_, counts: daskarr) -> daskarr:
    return counts


def norm_lib_size(assay, counts: daskarr) -> daskarr:
    return assay.sf * counts / assay.scalar.reshape(-1, 1)


def norm_lib_size_log(assay, counts: daskarr) -> daskarr:
    return np.log1p(assay.sf * counts / assay.scalar.reshape(-1, 1))


def norm_clr(_, counts: daskarr) -> daskarr:
    f = np.exp(np.log1p(counts).sum(axis=0) / len(counts))
    return np.log1p(counts / f)


def norm_tf_idf(assay, counts: daskarr) -> daskarr:
    tf = counts / assay.n_term_per_doc.reshape(-1, 1)
    # TODO: Split TF and IDF functionality to make it similar to norml_lib and zscaling
    idf = np.log2(1 + (assay.n_docs / (assay.n_docs_per_term + 1)))
    return tf * idf


class Assay:
    """
    A generic Assay class that contains methods to calculate feature level statistics.

    It also provides a method for saving normalized subset of data for later KNN graph construction.

    Attributes:
        name:
        z:
        cells:
        nthreads:
        rawData:
        feats: a MetaData object with info about each feature in the dataset
        attrs:
        normMethod: Which normalization method to use.
        sf: scaling factor for doing library-size normalization
    """
    def __init__(self, z: zarr.hierarchy, name: str, cell_data: MetaData,
                 nthreads: int, min_cells_per_feature: int = 10):
        """
        Args:
            z (zarr.hierarchy): Zarr hierarchy to use.
            name (str): Name for assay.
            cell_data: Metadata for the cells.
            nthreads:
            min_cells_per_feature:
        """
        self.name = name
        self.z = z[self.name]
        self.cells = cell_data
        self.nthreads = nthreads
        self.rawData = daskarr.from_zarr(self.z['counts'], inline_array=True)
        self.feats = MetaData(self.z['featureData'])
        self.attrs = self.z.attrs
        if 'percentFeatures' not in self.attrs:
            self.attrs['percentFeatures'] = {}
        self.normMethod = norm_dummy
        self.sf = None
        self._ini_feature_props(min_cells_per_feature)

    def normed(self, cell_idx: np.ndarray = None, feat_idx: np.ndarray = None, **kwargs):
        """

        Args:
            cell_idx:
            feat_idx:
            **kwargs:

        Returns:

        """
        if cell_idx is None:
            cell_idx = self.cells.active_index('I')
        if feat_idx is None:
            feat_idx = self.feats.active_index('I')
        counts = self.rawData[:, feat_idx][cell_idx, :]
        return self.normMethod(self, counts)

    def to_raw_sparse(self, cell_key):
        """

        Args:
            cell_key:

        Returns:

        """
        from tqdm import tqdm

        sm = None
        for i in tqdm(self.rawData[self.cells.active_index(cell_key), :].blocks, total=self.rawData.numblocks[0],
                      desc=f"INFO: Converting raw data from {self.name} assay into CSR format"):
            s = csr_matrix(controlled_compute(i, self.nthreads))
            if sm is None:
                sm = s
            else:
                sm = vstack([sm, s])
        return sm

    def _ini_feature_props(self, min_cells: int) -> None:
        """

        Args:
            min_cells:

        Returns:

        """
        if 'nCells' in self.feats.columns and 'dropOuts' in self.feats.columns:
            pass
        else:
            ncells = show_progress((self.rawData > 0).sum(axis=0),
                                   f"({self.name}) Computing nCells and dropOuts", self.nthreads)
            self.feats.insert('nCells', ncells, overwrite=True)
            self.feats.insert('dropOuts', abs(self.cells.N - self.feats.fetch('nCells')), overwrite=True)
            self.feats.update_key(ncells > min_cells, 'I')

    def add_percent_feature(self, feat_pattern: str, name: str) -> None:
        """

        Args:
            feat_pattern:
            name:

        Returns:

        """
        if name in self.attrs['percentFeatures']:
            if self.attrs['percentFeatures'][name] == feat_pattern:
                return None
            else:
                logger.info(f"Pattern for percentage feature {name} updated.")
        self.attrs['percentFeatures'] = {**{k: v for k, v in self.attrs['percentFeatures'].items()},
                                         **{name: feat_pattern}}
        feat_idx = sorted(self.feats.get_index_by(self.feats.grep(feat_pattern), 'names'))
        if len(feat_idx) == 0:
            logger.warning(f"No matches found for pattern {feat_pattern}."
                           f" Will not add/update percentage feature")
            return None
        total = show_progress(self.rawData[:, feat_idx].sum(axis=1),
                              f"Computing percentage of {name}", self.nthreads)
        if total.sum() == 0:
            logger.warning(f"Percentage feature {name} not added because not detected in any cell")
            return None
        self.cells.insert(name, 100 * total / self.cells.fetch_all(self.name + '_nCounts'), overwrite=True)

    def _verify_keys(self, cell_key: str, feat_key: str) -> None:
        if cell_key not in self.cells.columns or self.cells.get_dtype(cell_key) != bool:
            raise ValueError(f"ERROR: Either {cell_key} does not exist or is not bool type")
        if feat_key not in self.feats.columns or self.feats.get_dtype(feat_key) != bool:
            raise ValueError(f"ERROR: Either {feat_key} does not exist or is not bool type")

    def _get_cell_feat_idx(self, cell_key: str, feat_key: str) -> Tuple[np.ndarray, np.ndarray]:
        self._verify_keys(cell_key, feat_key)
        cell_idx = self.cells.active_index(cell_key)
        feat_idx = self.feats.active_index(feat_key)
        return cell_idx, feat_idx

    @staticmethod
    def _create_subset_hash(cell_idx: np.ndarray, feat_idx: np.ndarray) -> int:
        return hash(tuple([hash(tuple(cell_idx)), hash(tuple(feat_idx))]))

    @staticmethod
    def _get_summary_stats_loc(cell_key: str) -> Tuple[str, str]:
        return f"stats_{cell_key}", f"summary_stats_{cell_key}"

    def _validate_stats_loc(self, stats_loc: str, cell_idx: np.ndarray,
                            feat_idx: np.ndarray, delete_on_fail: bool = True) -> bool:
        subset_hash = self._create_subset_hash(cell_idx, feat_idx)
        if stats_loc in self.z:
            attrs = self.z[stats_loc].attrs
            if 'subset_hash' in attrs and attrs['subset_hash'] == subset_hash:
                return True
            else:
                # Reset stats loc
                if delete_on_fail:
                    del self.z[stats_loc]
                return False
        else:
            return False

    def _load_stats_loc(self, cell_key: str) -> str:
        cell_idx, feat_idx = self._get_cell_feat_idx(cell_key, 'I')
        identifier, stats_loc = self._get_summary_stats_loc(cell_key)
        if self._validate_stats_loc(stats_loc, cell_idx, feat_idx) is False:
            raise KeyError(f"Summary statistics have not been calculated for cell key: {cell_key}")
        if identifier not in self.feats.locations:
            self.feats.mount_location(self.z[stats_loc], identifier)
        else:
            logger.debug(f"Location ({stats_loc}) already mounted")
        return identifier

    def save_normalized_data(self, cell_key: str, feat_key: str, batch_size: int,
                             location: str, log_transform: bool, renormalize_subset: bool,
                             update_keys: bool) -> daskarr:
        """

        Args:
            cell_key:
            feat_key:
            batch_size:
            location:
            log_transform:
            renormalize_subset:
            update_keys:

        Returns:

        """

        from .writers import dask_to_zarr

        # FIXME: Extensive documentation needed to justify the naming strategy of slots here
        # Because HVGs and other feature selections have cell key appended in their metadata
        if feat_key != 'I':
            feat_key = cell_key + '__' + feat_key
        cell_idx, feat_idx = self._get_cell_feat_idx(cell_key, feat_key)
        subset_hash = self._create_subset_hash(cell_idx, feat_idx)
        subset_params = {'log_transform': log_transform, 'renormalize_subset': renormalize_subset}
        if location in self.z:
            if subset_hash == self.z[location].attrs['subset_hash'] and \
                    subset_params == self.z[location].attrs['subset_params']:
                logger.info(f"Using existing normalized data with cell key {cell_key} and feat key {feat_key}")
                if update_keys:
                    self.attrs['latest_feat_key'] = feat_key.split('__', 1)[1] if feat_key != 'I' else 'I'
                    self.attrs['latest_cell_key'] = cell_key
                return daskarr.from_zarr(self.z[location + '/data'], inline_array=True)
            else:
                # Creating group here to overwrite all children
                self.z.create_group(location, overwrite=True)
        vals = self.normed(cell_idx, feat_idx, log_transform=log_transform,
                           renormalize_subset=renormalize_subset)
        dask_to_zarr(vals, self.z, location + '/data', batch_size, self.nthreads)
        self.z[location].attrs['subset_hash'] = subset_hash
        self.z[location].attrs['subset_params'] = subset_params
        if update_keys:
            self.attrs['latest_feat_key'] = feat_key.split('__', 1)[1] if feat_key != 'I' else 'I'
            self.attrs['latest_cell_key'] = cell_key
        return daskarr.from_zarr(self.z[location + '/data'], inline_array=True)

    def score_features(self, feature_names: List[str], cell_key: str,
                       ctrl_size: int, n_bins: int, rand_seed: int) -> np.ndarray:
        """

        Args:
            feature_names:
            cell_key:
            ctrl_size:
            n_bins:
            rand_seed:

        Returns:

        """

        from .feat_utils import binned_sampling

        def _names_to_idx(i):
            return self.feats.get_index_by(i, 'names', None)

        def _calc_mean(i):
            return self.normed(cell_idx=cell_idx, feat_idx=np.array(sorted(i))).mean(axis=1).compute()

        feature_idx = _names_to_idx(feature_names)
        if len(feature_idx) == 0:
            raise ValueError(f"ERROR: No feature ids found for any of the provided {len(feature_names)} features")

        identifier = self._load_stats_loc(cell_key)
        obs_avg = pd.Series(self.feats.fetch_all(f"{identifier}_avg"))
        control_idx = binned_sampling(obs_avg, list(feature_idx), ctrl_size, n_bins, rand_seed)
        cell_idx, _ = self._get_cell_feat_idx(cell_key, 'I')
        return _calc_mean(feature_idx) - _calc_mean(control_idx)

    def __repr__(self):
        f = self.feats.fetch_all('I')
        assay_name = str(self.__class__).split('.')[-1][:-2]
        return f"{assay_name} {self.name} with {f.sum()}({len(f)}) features"


class RNAassay(Assay):
    """
    This assay is designed for feature selection and normalization of scRNA-Seq data.

    Subclass of Assay.
    """
    def __init__(self, z: zarr.hierarchy, name: str, cell_data: MetaData, **kwargs):
        """

        Args:
            z:
            name:
            cell_data:
            **kwargs:
        """
        super().__init__(z, name, cell_data, **kwargs)
        self.normMethod = norm_lib_size
        if 'size_factor' in self.attrs:
            self.sf = int(self.attrs['size_factor'])
        else:
            self.sf = 1000
            self.attrs['size_factor'] = self.sf
        self.scalar = None

    def normed(self, cell_idx: np.ndarray = None, feat_idx: np.ndarray = None,
               renormalize_subset: bool = False, log_transform: bool = False, **kwargs):
        """
        Args:
            cell_idx:
            feat_idx:
            renormalize_subset:
            log_transform:
            **kwargs:

        Returns:

        """
        if cell_idx is None:
            cell_idx = self.cells.active_index('I')
        if feat_idx is None:
            feat_idx = self.feats.active_index('I')
        counts = self.rawData[:, feat_idx][cell_idx, :]
        norm_method_cache = self.normMethod
        if log_transform:
            self.normMethod = norm_lib_size_log
        if renormalize_subset:
            a = show_progress(counts.sum(axis=1), "Normalizing with feature subset", self.nthreads)
            a[a == 0] = 1
            self.scalar = a
        else:
            self.scalar = self.cells.fetch_all(self.name + '_nCounts')[cell_idx]
        val = self.normMethod(self, counts)
        self.normMethod = norm_method_cache
        return val

    def set_feature_stats(self, cell_key: str, min_cells: int) -> None:
        """

        Args:
            cell_key:
            min_cells:

        Returns:

        """
        feat_key = 'I'  # Here we choose to calculate stats for all the features
        cell_idx, feat_idx = self._get_cell_feat_idx(cell_key, feat_key)
        identifier, stats_loc = self._get_summary_stats_loc(cell_key)
        if self._validate_stats_loc(stats_loc, cell_idx, feat_idx) is True:
            logger.info(f"Using cached feature stats for cell_key {cell_key}")
            return None
        n_cells = show_progress((self.normed(cell_idx, feat_idx) > 0).sum(axis=0),
                                f"({self.name}) Computing nCells", self.nthreads)
        tot = show_progress(self.normed(cell_idx, feat_idx).sum(axis=0),
                            f"({self.name}) Computing normed_tot", self.nthreads)
        sigmas = show_progress(self.normed(cell_idx, feat_idx).var(axis=0),
                               f"({self.name}) Computing sigmas", self.nthreads)
        idx = n_cells > min_cells
        self.feats.update_key(idx, key=feat_key)
        n_cells, tot, sigmas = n_cells[idx], tot[idx], sigmas[idx]

        self.z.create_group(stats_loc, overwrite=True)
        self.feats.mount_location(self.z[stats_loc], identifier)
        self.feats.insert('normed_tot', tot.astype(float), overwrite=True, location=identifier)
        self.feats.insert('avg', (tot / self.cells.N).astype(float), overwrite=True, location=identifier)
        self.feats.insert('nz_mean', (tot / n_cells).astype(float), overwrite=True, location=identifier)
        self.feats.insert('sigmas', sigmas.astype(float), overwrite=True, location=identifier)
        self.feats.insert('normed_n', n_cells.astype(float), overwrite=True, location=identifier)
        self.z[stats_loc].attrs['subset_hash'] = self._create_subset_hash(
            cell_idx, self.feats.active_index(feat_key))
        self.feats.unmount_location(identifier)
        return None

    # maybe we should return plot here? If one wants to modify it. /raz
    def mark_hvgs(self, cell_key: str, min_cells: int, top_n: int,
                  min_var: float, max_var: float, min_mean: float, max_mean: float,
                  n_bins: int, lowess_frac: float, blacklist: str, hvg_key_name: str,
                  show_plot: bool, **plot_kwargs) -> None:
        """
        Identifies highly variable genes in the dataset.
        
        The parameters govern the min/max variance (corrected) and mean expression threshold for calling genes highly
        variable. The variance is corrected by first dividing genes into bins based on their mean expression values.
        Genes with minimum variance is selected from each bin and a Lowess curve is fitted to
        the mean-variance trend of these genes. mark_hvgs will by default run on the default assay.

        A plot is produced, that for each gene shows the corrected variance on the y-axis and the non-zero mean
        (means from cells where the gene had a non-zero value) on the x-axis. The genes are colored in two gradients
        which indicate the number of cells where the gene was expressed. The colors are yellow to dark red for HVGs,
        and blue to green for non-HVGs.

        The mark_hvgs function has a parameter cell_key that dictates which cells to use to identify the HVGs.
        The default value of this parameter is I, which means it will use all the cells that were not filtered out.
        
        *Modifies the feats table*: adds a column named `<cell_key>__hvgs` to the feature table,
        which contains a True value for genes marked HVGs. The prefix comes from the `cell_key` parameter,
        the naming rule in Scarf dictates that cells used to identify HVGs are prepended to the column name
        (with a double underscore delimiter).

        Args:
            cell_key: Specify which cells to use to identify the HVGs. Default value `I` (use all non-filtered out cells).
            min_cells:
            top_n:
            min_var:
            max_var:
            min_mean:
            max_mean:
            n_bins: Number of bins
            lowess_frac:
            blacklist:
            hvg_key_name:
            show_plot:
            **plot_kwargs:
        """
        self.set_feature_stats(cell_key, min_cells)
        identifier = self._load_stats_loc(cell_key)
        col_renamer = lambda x: f"{identifier}_{x}"
        c_var_col = f"c_var__{n_bins}__{lowess_frac}"
        if col_renamer(c_var_col) in self.feats.columns:
            logger.info("Using existing corrected dispersion values")
        else:
            slots = ['normed_tot', 'avg', 'nz_mean', 'sigmas', 'normed_n']
            for i in slots:
                i = col_renamer(i)
                if i not in self.feats.columns:
                    raise KeyError("ERROR: {i} not found in feature metadata")
            c_var = self.feats.remove_trend(col_renamer('avg'), col_renamer('sigmas'),
                                            n_bins, lowess_frac)
            self.feats.insert(c_var_col, c_var, overwrite=True, location=identifier)

        if max_mean != np.Inf:
            max_mean = 2 ** max_mean
        if max_var != np.Inf:
            max_var = 2 ** max_var
        if min_mean != -np.Inf:
            min_mean = 2 ** min_mean
        if min_var != -np.Inf:
            min_var = 2 ** min_var

        bl = self.feats.index_to_bool(self.feats.get_index_by(self.feats.grep(blacklist), 'names'), invert=True)
        if min_var == -np.Inf:
            if top_n < 1:
                raise ValueError("ERROR: Please provide a value greater than 0 for `top_n` parameter")
            idx = self.feats.multi_sift(
                [col_renamer('normed_n'), col_renamer('nz_mean')], [min_cells, min_mean], [np.Inf, max_mean])
            idx = idx & self.feats.fetch_all('I') & bl
            n_valid_feats = idx.sum()
            if top_n > n_valid_feats:
                logger.warning(f"WARNING: Number of valid features are less then value "
                               f"of parameter `top_n`: {top_n}. Resetting `top_n` to {n_valid_feats}")
                top_n = n_valid_feats - 1
            min_var = pd.Series(self.feats.fetch_all(col_renamer(c_var_col))
                                )[idx].sort_values(ascending=False).values[top_n]
        hvgs = self.feats.multi_sift(
            [col_renamer(x) for x in ['normed_n', 'nz_mean', c_var_col]],
            [min_cells, min_mean, min_var], [np.Inf, max_mean, max_var])
        hvgs = hvgs & self.feats.fetch_all('I') & bl
        hvg_key_name = cell_key + '__' + hvg_key_name
        logger.info(f"{sum(hvgs)} genes marked as HVGs")
        self.feats.insert(hvg_key_name, hvgs, fill_value=False, overwrite=True)

        if show_plot:
            from .plots import plot_mean_var
            nzm, vf, nc = [self.feats.fetch(x) for x in [col_renamer('nz_mean'), col_renamer(c_var_col), 'nCells']]
            plot_mean_var(nzm, vf, nc, self.feats.fetch(hvg_key_name), **plot_kwargs)

        return None


class ATACassay(Assay):
    # TODO: add docstring
    def __init__(self, z: zarr.hierarchy, name: str, cell_data: MetaData, **kwargs):
        """

        Args:
            z:
            name:
            cell_data:
            **kwargs:
        """
        super().__init__(z, name, cell_data, **kwargs)
        self.normMethod = norm_tf_idf
        self.n_term_per_doc = None
        self.n_docs = None
        self.n_docs_per_term = None

    def normed(self, cell_idx: np.ndarray = None, feat_idx: np.ndarray = None, **kwargs):
        """

        Args:
            cell_idx:
            feat_idx:
            **kwargs:

        Returns:

        """
        if cell_idx is None:
            cell_idx = self.cells.active_index('I')
        if feat_idx is None:
            feat_idx = self.feats.active_index('I')
        counts = self.rawData[:, feat_idx][cell_idx, :]
        self.n_term_per_doc = self.cells.fetch_all(self.name + '_nFeatures')[cell_idx]
        self.n_docs = len(cell_idx)
        self.n_docs_per_term = self.feats.fetch_all('nCells')[feat_idx]
        return self.normMethod(self, counts)

    def set_feature_stats(self, cell_key: str) -> None:
        """

        Args:
            cell_key:

        Returns:

        """
        feat_key = 'I'  # Here we choose to calculate stats for all the features
        cell_idx, feat_idx = self._get_cell_feat_idx(cell_key, feat_key)
        identifier, stats_loc = self._get_summary_stats_loc(cell_key)
        if self._validate_stats_loc(stats_loc, cell_idx, feat_idx) is True:
            logger.info(f"Using cached feature stats for cell_key {cell_key}")
            return None
        prevalence = show_progress(self.normed(cell_idx, feat_idx).sum(axis=0),
                                   f"({self.name}) Calculating peak prevalence across cells", self.nthreads)
        self.z.create_group(stats_loc, overwrite=True)
        self.feats.mount_location(self.z[stats_loc], identifier)
        self.feats.insert('prevalence', prevalence.astype(float), overwrite=True, location=identifier)
        self.z[stats_loc].attrs['subset_hash'] = self._create_subset_hash(cell_idx, feat_idx)
        self.feats.unmount_location(identifier)
        return None

    def mark_prevalent_peaks(self, cell_key: str, top_n: int, prevalence_key_name: str) -> None:
        """

        Args:
            cell_key:
            top_n:
            prevalence_key_name:

        Returns:

        """
        if top_n >= self.feats.N:
            raise ValueError(f"ERROR: n_top should be less than total number of features ({self.feats.N})]")
        if type(top_n) != int:
            raise TypeError("ERROR: n_top must a positive integer value")
        self.set_feature_stats(cell_key)
        identifier = self._load_stats_loc(cell_key)
        col_renamer = lambda x: f"{identifier}_{x}"
        idx = pd.Series(self.feats.fetch_all(col_renamer('prevalence'))).sort_values(ascending=False)[:top_n].index
        prevalence_key_name = cell_key + '__' + prevalence_key_name
        self.feats.insert(prevalence_key_name, self.feats.index_to_bool(idx), fill_value=False, overwrite=True)
        return None


class ADTassay(Assay):
    # TODO: add docstring
    def __init__(self, z: zarr.hierarchy, name: str, cell_data: MetaData, **kwargs):
        """

        Args:
            z:
            name:
            cell_data:
            **kwargs:
        """
        super().__init__(z, name, cell_data, **kwargs)
        self.normMethod = norm_clr

    def normed(self, cell_idx: np.ndarray = None, feat_idx: np.ndarray = None, **kwargs):
        """

        Args:
            cell_idx:
            feat_idx:
            **kwargs:

        Returns:

        """
        if cell_idx is None:
            cell_idx = self.cells.active_index('I')
        if feat_idx is None:
            feat_idx = self.feats.active_index('I')
        counts = self.rawData[:, feat_idx][cell_idx, :]
        return self.normMethod(self, counts)



"""
Module to find biomarkers.
"""
from .assay import Assay
from .utils import controlled_compute
from numba import jit
import numpy as np
import pandas as pd
from tqdm import tqdm


__all__ = ['find_markers_by_rank']


def find_markers_by_rank(assay: Assay, group_key: str, cell_key: str,
                         nthreads: int, threshold: float = 0.25, gene_batch_size: int = 50) -> dict:
    """

    Args:
        assay:
        group_key:
        cell_key:
        nthreads:
        threshold:
        gene_batch_size:

    Returns:

    """

    @jit(nopython=True)
    def calc_mean_rank(v):
        """
        Calculates the mean rank of the data.
        """
        r = np.ones(n_groups)
        for x in range(n_groups):
            r[x] = v[int_indices == x].mean()
        return r / r.sum()

    def mean_rank_wrapper(v):
        """
        Wraps `calc_mean_rank` function.
        """
        return calc_mean_rank(v.values)

    groups = assay.cells.fetch(group_key, cell_key)
    group_set = sorted(set(groups))
    n_groups = len(group_set)
    # Since, numba needs int arrays to work properly but the dtype of 'groups' may not be integer type
    # Hence we need to create a indexed version of 'groups'
    idx_map = dict(zip(group_set, range(n_groups)))
    rev_idx_map = {v: k for k, v in idx_map.items()}
    int_indices = np.array([idx_map[x] for x in groups])

    data = assay.normed(cell_idx=assay.cells.active_index(cell_key))
    gene_ids = assay.feats.fetch('ids')
    chunks = np.array_split(np.arange(0, data.shape[1]), int(data.shape[1]/gene_batch_size))

    results = {x: [] for x in group_set}
    for chunk in tqdm(chunks, desc='Finding markers', total=len(chunks)):
        val = pd.DataFrame(controlled_compute(data[:, chunk], nthreads),
                           columns=gene_ids[chunk]).rank(method='dense').astype(int)
        res = val.apply(mean_rank_wrapper)
        # Removing genes that were below the threshold in all the groups
        res = res.T[(res < threshold).sum() != n_groups]
        for j in res:
            results[rev_idx_map[j]].append(res[j][res[j] > threshold])
    for i in results:
        results[i] = pd.concat(results[i]).sort_values(ascending=False)
    return results



"""
Methods and classes for writing data to disk.

- Methods:
    - create_zarr_dataset: Creates and returns a Zarr hierarchy/dataset.
    - create_zarr_obj_array: Creates and returns a Zarr object array.
    - create_zarr_count_assay: Creates and returns a Zarr array with name 'counts'.
    - subset_assay_zarr: Selects a subset of the data in an assay in the specified Zarr hierarchy.
    - dask_to_zarr: Creates a Zarr hierarchy from a Dask array.

- Classes:
    - ZarrMerge: Merge multiple Zarr files into a single Zarr file.
    - CrToZarr: A class for converting data in the Cellranger format to a Zarr hierarchy.
    - H5adToZarr: A class for converting data in the Cellranger Matrix Market format to a Zarr hierarchy.
    - MtxToZarr: A class for converting data in the Cellranger Matrix Market format to a Zarr hierarchy.
    - NaboH5ToZarr: A class for converting data in a h5 file generated by Nabo, to a Zarr hierarchy.
    - LoomToZarr: A class for converting data in a Loom file to a Zarr hierarchy.
"""

import zarr
from typing import Any, Tuple, List, Union
import numpy as np
from tqdm import tqdm
from .readers import CrReader, H5adReader, NaboH5Reader, LoomReader
import os
import pandas as pd
from .utils import controlled_compute
from .logging_utils import logger
from scipy.sparse import csr_matrix

__all__ = ['create_zarr_dataset', 'create_zarr_obj_array', 'create_zarr_count_assay',
           'subset_assay_zarr', 'dask_to_zarr', 'ZarrMerge', 'SubsetZarr',
           'CrToZarr', 'H5adToZarr', 'MtxToZarr', 'NaboH5ToZarr', 'LoomToZarr', 'SparseToZarr']


def create_zarr_dataset(g: zarr.Group, name: str, chunks: tuple,
                        dtype: Any, shape: Tuple, overwrite: bool = True) -> zarr.hierarchy:
    """
    Creates and returns a Zarr array.

    Args:
        g (zarr.hierarchy):
        name (str):
        chunks (tuple):
        dtype (Any):
        shape (Tuple):
        overwrite (bool):

    Returns:
        A Zarr Array.
    """
    from numcodecs import Blosc

    compressor = Blosc(cname='lz4', clevel=5, shuffle=Blosc.BITSHUFFLE)
    return g.create_dataset(name, chunks=chunks, dtype=dtype,
                            shape=shape, compressor=compressor, overwrite=overwrite)


def create_zarr_obj_array(g: zarr.Group, name: str, data,
                          dtype: Union[str, Any] = None, overwrite: bool = True) -> zarr.hierarchy:
    """
    Creates and returns a Zarr object array.

    A Zarr object array can contain any type of object.
    https://zarr.readthedocs.io/en/stable/tutorial.html#object-arrays

    Args:
        g (zarr.hierarchy):
        name (str):
        data ():
        dtype (Union[str, Any]):
        overwrite (bool):

    Returns:
        A Zarr object Array.
    """
    data = np.array(data)
    if dtype is None or dtype == object:
        dtype = 'U' + str(max([len(str(x)) for x in data]))
    if np.issubdtype(data.dtype, np.dtype('S')):
        data = data.astype('U')
        dtype = data.dtype
    return g.create_dataset(name, data=data, chunks=(100000,),
                            shape=len(data), dtype=dtype, overwrite=overwrite)


def create_zarr_count_assay(z: zarr.Group, assay_name: str, chunk_size: Tuple[int, int], n_cells: int,
                            feat_ids: Union[np.ndarray, List[str]], feat_names: Union[np.ndarray, List[str]],
                            dtype: str = 'uint32') -> zarr.hierarchy:
    """
    Creates and returns a Zarr array with name 'counts'.

    Args:
        z (zarr.Group):
        assay_name (str):
        chunk_size (Tuple[int, int]):
        n_cells (int):
        feat_ids (Union[np.ndarray, List[str]]):
        feat_names (Union[np.ndarray, List[str]]):
        dtype (str = 'uint32'):

    Returns:
        A Zarr array.
    """
    g = z.create_group(assay_name, overwrite=True)
    g.attrs['is_assay'] = True
    g.attrs['misc'] = {}
    create_zarr_obj_array(g, 'featureData/ids', feat_ids)
    create_zarr_obj_array(g, 'featureData/names', feat_names)
    create_zarr_obj_array(g, 'featureData/I',
                          [True for _ in range(len(feat_ids))], 'bool')
    return create_zarr_dataset(g, 'counts', chunk_size, dtype,
                               (n_cells, len(feat_ids)), overwrite=True)


class CrToZarr:
    """
    A class for converting data in the Cellranger format to a Zarr hierarchy.

    Attributes:
        cr: A CrReader object, containing the Cellranger data.
        fn: The file name for the Zarr hierarchy.
        chunkSizes: The requested size of chunks to load into memory and process.
        z: The Zarr hierarchy (array or group).
    """
    def __init__(self, cr: CrReader, zarr_fn: str, chunk_size=(1000, 1000), dtype: str = 'uint32'):
        """
        Args:
            cr: A CrReader object, containing the Cellranger data.
            zarr_fn: The file name for the Zarr hierarchy.
            chunk_size: The requested size of chunks to load into memory and process.
            dtype: the dtype of the data.
        """
        self.cr = cr
        self.fn = zarr_fn
        self.chunkSizes = chunk_size
        self.z = zarr.open(self.fn, mode='w')
        self._ini_cell_data()
        for assay_name in self.cr.assayFeats.columns:
            create_zarr_count_assay(self.z, assay_name, chunk_size, self.cr.nCells,
                                    self.cr.feature_ids(assay_name), self.cr.feature_names(assay_name), dtype)

    def _ini_cell_data(self):
        g = self.z.create_group('cellData')
        create_zarr_obj_array(g, 'ids', self.cr.cell_names())
        create_zarr_obj_array(g, 'names', self.cr.cell_names())
        create_zarr_obj_array(g, 'I', [True for _ in range(self.cr.nCells)], 'bool')

    def dump(self, batch_size: int = 1000, lines_in_mem: int = 100000) -> None:
        # TODO: add informed description to docstring
        """
        Raises:
            AssertionError: Catches eventual bugs in the class, if number of cells does not match after transformation.

        Returns:
            None
        """
        stores = [self.z["%s/counts" % x] for x in self.cr.assayFeats.columns]
        spidx = [0] + list(self.cr.assayFeats.T.nFeatures.cumsum().values)
        spidx = [(spidx[x - 1], spidx[x]) for x in range(1, len(spidx))]
        s, e, = 0, 0
        n_chunks = self.cr.nCells//batch_size + 1
        for a in tqdm(self.cr.consume(batch_size, lines_in_mem), total=n_chunks):
            e += a.shape[0]
            a = a.todense()
            for j in range(len(stores)):
                stores[j][s:e] = a[:, spidx[j][0]:spidx[j][1]]
            s = e
        if e != self.cr.nCells:
            raise AssertionError("ERROR: This is a bug in CrToZarr. All cells might not have been successfully "
                                 "written into the zarr file. Please report this issue")


class MtxToZarr:
    """
    A class for converting data in the Cellranger Matrix Market format to a Zarr hierarchy.

    Attributes:
        cr: A CrReader object, containing the Cellranger data.
        fn: The file name for the Zarr hierarchy.
        chunkSizes: The requested size of chunks to load into memory and process.
        z: The Zarr hierarchy (array or group).
    """
    def __init__(self, cr: CrReader, zarr_fn: str, chunk_size=(1000, 1000), dtype: str = 'uint32'):
        """
        Args:
            cr: A CrReader object, containing the Cellranger data.
            zarr_fn: The file name for the Zarr hierarchy.
            chunk_size: The requested size of chunks to load into memory and process.
            dtype: the dtype of the data.
        """
        self.cr = cr
        self.fn = zarr_fn
        self.chunkSizes = chunk_size
        self.z = zarr.open(self.fn, mode='w')
        self._ini_cell_data()
        for assay_name in set(self.cr.assayFeats.columns):
            create_zarr_count_assay(self.z, assay_name, chunk_size, self.cr.nCells,
                                    self.cr.feature_ids(assay_name), self.cr.feature_names(assay_name), dtype)

    def _ini_cell_data(self):
        g = self.z.create_group('cellData')
        create_zarr_obj_array(g, 'ids', self.cr.cell_names())
        create_zarr_obj_array(g, 'names', self.cr.cell_names())
        create_zarr_obj_array(g, 'I', [True for _ in range(self.cr.nCells)], 'bool')

    def _prep_assay_ranges(self):
        ret_val = {}
        for assay in set(self.cr.assayFeats.columns):
            temp = []
            if len(self.cr.assayFeats[assay].shape) == 2:
                for i in self.cr.assayFeats[assay].values[1:3].T:
                    temp.append([i[0], i[1]])
            else:
                idx = self.cr.assayFeats[assay]
                temp = [[idx.start, idx.end]]
            ret_val[assay] = temp
        return ret_val

    def dump(self, batch_size: int = 1000, lines_in_mem: int = 100000) -> None:
        # TODO: add informed description to docstring
        """
        Raises:
            AssertionError: Catches eventual bugs in the class, if number of cells does not match after transformation.

        Returns:
            None
        """
        stores = {x: self.z["%s/counts" % x] for x in set(self.cr.assayFeats.columns)}
        assay_ranges = self._prep_assay_ranges()
        s, e, = 0, 0
        n_chunks = self.cr.nCells // batch_size + 1
        for a in tqdm(self.cr.consume(batch_size, lines_in_mem), total=n_chunks):
            e += a.shape[0]
            a = a.todense()
            b = {x: [] for x in stores.keys()}
            for store_name in stores.keys():
                ar = assay_ranges[store_name]
                temp = []
                for i in ar:
                    temp.append(a[:, i[0]:i[1]])
                if len(temp) > 1:
                    b[store_name] = np.hstack(temp)
                else:
                    b[store_name] = temp[0]
            for store_name in stores.keys():
                stores[store_name][s:e] = b[store_name]
            s = e
        if e != self.cr.nCells:
            raise AssertionError("ERROR: This is a bug in MtxToZarr. All cells might not have been successfully "
                                 "written into the zarr file. Please report this issue")


class H5adToZarr:
    """
    A class for converting data in the Cellranger Matrix Market format to a Zarr hierarchy.

    Attributes:
        h5ad: A h5ad object (h5 file with added AnnData structure).
        fn: The file name for the Zarr hierarchy.
        chunkSizes: The requested size of chunks to load into memory and process.
        assayName: The Zarr hierarchy (array or group).
        z: The Zarr hierarchy (array or group).
    """
    def __init__(self, h5ad: H5adReader, zarr_fn: str, assay_name: str = None,
                 chunk_size=(1000, 1000)):
        """
        Args:
            h5ad: A H5adReader object, containing the Cellranger data.
            zarr_fn: The file name for the Zarr hierarchy.
            assay_name: the name of the assay (e. g. 'RNA')
            chunk_size: The requested size of chunks to load into memory and process.
        """
        # TODO: support for multiple assay. One of the `var` datasets can be used to group features in separate assays
        self.h5ad = h5ad
        self.fn = zarr_fn
        self.chunkSizes = chunk_size
        if assay_name is None:
            logger.info(f"No value provided for assay names. Will use default value: 'RNA'")
            self.assayName = 'RNA'
        else:
            self.assayName = assay_name
        self.z = zarr.open(self.fn, mode='w')
        self._ini_cell_data()
        create_zarr_count_assay(self.z, self.assayName, chunk_size, self.h5ad.nCells,
                                self.h5ad.feat_ids(), self.h5ad.feat_names(), self.h5ad.matrixDtype)
        for i, j in self.h5ad.get_feat_columns():
            if i not in self.z[self.assayName]['featureData']:
                create_zarr_obj_array(self.z[self.assayName]['featureData'], i, j, j.dtype)

    def _ini_cell_data(self):
        g = self.z.create_group('cellData')
        ids = self.h5ad.cell_ids()
        create_zarr_obj_array(g, 'ids', ids, ids.dtype)
        create_zarr_obj_array(g, 'names', ids, ids.dtype)
        create_zarr_obj_array(g, 'I', [True for _ in range(self.h5ad.nCells)], 'bool')
        for i, j in self.h5ad.get_cell_columns():
            create_zarr_obj_array(g, i, j, j.dtype)

    def dump(self, batch_size: int = 1000) -> None:
        # TODO: add informed description to docstring
        """
        Raises:
            AssertionError: Catches eventual bugs in the class, if number of cells does not match after transformation.

        Returns:
            None
        """
        store = self.z["%s/counts" % self.assayName]
        s, e, = 0, 0
        n_chunks = self.h5ad.nCells//batch_size + 1
        for a in tqdm(self.h5ad.consume(batch_size), total=n_chunks):
            e += a.shape[0]
            store[s:e] = a
            s = e
        if e != self.h5ad.nCells:
            raise AssertionError("ERROR: This is a bug in H5adToZarr. All cells might not have been successfully "
                                 "written into the zarr file. Please report this issue")


class NaboH5ToZarr:
    """
    A class for converting data in a h5 file generated by Nabo, to a Zarr hierarchy.

    Attributes:
        h5: A Nabo h5 object.
        fn: The file name for the Zarr hierarchy.
        chunkSizes: The requested size of chunks to load into memory and process.
        assayName: The Zarr hierarchy (array or group).
        z: The Zarr hierarchy (array or group).
    """
    def __init__(self, h5: NaboH5Reader, zarr_fn: str, assay_name: str = None,
                 chunk_size=(1000, 1000), dtype: str = 'uint32'):
        """
        Args:
            h5: A Nabo h5 object containing the data.
            zarr_fn: The file name for the Zarr hierarchy.
            assay_name: the name of the assay (e. g. 'RNA')
            chunk_size: The requested size of chunks to load into memory and process.
            dtype: the dtype of the data.
        """
        self.h5 = h5
        self.fn = zarr_fn
        self.chunkSizes = chunk_size
        if assay_name is None:
            logger.info(f"No value provided for assay names. Will use default value: 'RNA'")
            self.assayName = 'RNA'
        else:
            self.assayName = assay_name
        self.z = zarr.open(self.fn, mode='w')
        self._ini_cell_data()
        create_zarr_count_assay(self.z, self.assayName, chunk_size, self.h5.nCells,
                                self.h5.feat_ids(), self.h5.feat_names(), dtype)

    def _ini_cell_data(self):
        g = self.z.create_group('cellData')
        create_zarr_obj_array(g, 'ids', self.h5.cell_ids())
        create_zarr_obj_array(g, 'names', self.h5.cell_ids())
        create_zarr_obj_array(g, 'I', [True for _ in range(self.h5.nCells)], 'bool')

    def dump(self, batch_size: int = 500) -> None:
        # TODO: add informed description to docstring
        """
        Raises:
            AssertionError: Catches eventual bugs in the class, if number of cells does not match after transformation.

        Returns:
            None
        """
        store = self.z["%s/counts" % self.assayName]
        s, e, = 0, 0
        n_chunks = self.h5.nCells // batch_size + 1
        for a in tqdm(self.h5.consume(batch_size), total=n_chunks):
            e += a.shape[0]
            store[s:e] = a
            s = e
        if e != self.h5.nCells:
            raise AssertionError("ERROR: This is a bug in NaboH5ToZarr. All cells might not have been successfully "
                                 "written into the zarr file. Please report this issue")


class LoomToZarr:
    """
    A class for converting data in a Loom file to a Zarr hierarchy.

    Converts a Loom file read using scarf.LoomReader into Scarf's Zarr format.
    
    Attributes:
        loom: A scarf.LoomReader object used to open Loom format file.
        fn: The file name for the Zarr hierarchy.
        chunkSizes: The requested size of chunks to load into memory and process.
        assayName: The Zarr hierarchy (array or group).
        z: The Zarr hierarchy (array or group).
    """
    def __init__(self, loom: LoomReader, zarr_fn: str, assay_name: str = None,
                 chunk_size=(1000, 1000)):
        """
        Args:
            loom: LoomReader object used to open Loom format file
            zarr_fn: Output Zarr filename with path
            assay_name: Name for the output assay. If not provided then automatically set to RNA
            chunk_size: Chunk size for the count matrix saved in Zarr file.
        """
        # TODO: support for multiple assay. Data from within individual layers can be treated as separate assays
        self.loom = loom
        self.fn = zarr_fn
        self.chunkSizes = chunk_size
        if assay_name is None:
            logger.info(f"No value provided for assay names. Will use default value: 'RNA'")
            self.assayName = 'RNA'
        else:
            self.assayName = assay_name
        self.z = zarr.open(self.fn, mode='w')
        self._ini_cell_data()
        create_zarr_count_assay(self.z, self.assayName, chunk_size, self.loom.nCells,
                                self.loom.feature_ids(), self.loom.feature_names(), self.loom.matrixDtype)
        for i, j in self.loom.get_feature_attrs():
            create_zarr_obj_array(self.z[self.assayName]['featureData'], i, j, j.dtype)

    def _ini_cell_data(self):
        g = self.z.create_group('cellData')
        create_zarr_obj_array(g, 'ids', self.loom.cell_ids())
        create_zarr_obj_array(g, 'names', self.loom.cell_ids())
        create_zarr_obj_array(g, 'I', [True for _ in range(self.loom.nCells)], 'bool')
        for i, j in self.loom.get_cell_attrs():
            create_zarr_obj_array(g, i, j, j.dtype)

    def dump(self, batch_size: int = 1000) -> None:
        # TODO: add informed description to docstring
        """
        Raises:
            AssertionError: Catches eventual bugs in the class, if number of cells does not match after transformation.

        Returns:
            None
        """
        store = self.z["%s/counts" % self.assayName]
        s, e, = 0, 0
        n_chunks = self.loom.nCells//batch_size + 1
        for a in tqdm(self.loom.consume(batch_size), total=n_chunks):
            e += a.shape[0]
            store[s:e] = a
            s = e
        if e != self.loom.nCells:
            raise AssertionError("ERROR: This is a bug in LoomToZarr. All cells might not have been successfully "
                                 "written into the zarr file. Please report this issue")


class SparseToZarr:
    """
    A class for converting data in a sparse matrix to a Zarr hierarchy.

    Converts a Loom file read using scarf.LoomReader into Scarf's Zarr format.

    Attributes:
        csr_mat:
        fn: The file name for the Zarr hierarchy.
        chunkSizes: The requested size of chunks to load into memory and process.
        assayName: The Zarr hierarchy (array or group).
        z: The Zarr hierarchy (array or group).
    """

    def __init__(self, csr_mat: csr_matrix, zarr_fn: str, cell_ids: List[str], feature_ids: List[str],
                 assay_name: str = None, chunk_size=(1000, 1000), ):
        """
        Args:
            csr_mat:
            zarr_fn: Output Zarr filename with path
            cell_ids: Cell IDs for the cells in the dataset.
            feature_ids: Feature IDs for the features in the dataset.
            assay_name: Name for the output assay. If not provided then automatically set to RNA.
            chunk_size: The requested size of chunks to load into memory and process.

        Raises:
            ValueError: Raised if number of input cell or feature IDs does not match the matrix.
            AssertionError: Catches eventual bugs in the class, if number of cells does not match after transformation.
        """
        self.mat = csr_mat
        self.fn = zarr_fn
        self.chunkSizes = chunk_size
        if assay_name is None:
            logger.info(f"No value provided for assay names. Will use default value: 'RNA'")
            self.assayName = 'RNA'
        else:
            self.assayName = assay_name
        self.nFeatures, self.nCells = self.mat.shape
        if len(cell_ids) != self.nCells:
            raise ValueError("ERROR: Number of cell ids are not same as number of cells in the matrix")
        if len(feature_ids) != self.nFeatures:
            raise ValueError("ERROR: Number of feature ids are not same as number of features in the matrix")

        self.z = zarr.open(self.fn, mode='w')
        self._ini_cell_data(cell_ids)
        create_zarr_count_assay(self.z, self.assayName, chunk_size, self.nCells,
                                feature_ids, feature_ids, 'int64')

    def _ini_cell_data(self, cell_ids):
        g = self.z.create_group('cellData')
        create_zarr_obj_array(g, 'ids', cell_ids)
        create_zarr_obj_array(g, 'names', cell_ids)
        create_zarr_obj_array(g, 'I', [True for _ in range(self.nCells)], 'bool')

    def dump(self, batch_size: int = 1000) -> None:
        # TODO: add informed description to docstring
        """
        Raises:
            ValueError: Raised if there is any unexpected errors when writing to the Zarr hierarchy.
            AssertionError: Catches eventual bugs in the class, if number of cells does not match after transformation.

        Returns:
            None
        """
        store = self.z["%s/counts" % self.assayName]
        s, e, = 0, 0
        n_chunks = self.nCells//batch_size + 1
        for e in tqdm(range(batch_size, self.nCells+batch_size, batch_size), total=n_chunks):
            if s == self.nCells:
                raise ValueError("Unexpected error encountered in writing to Zarr. The last iteration has failed. "
                                 "Please report this issue.")
            if e > self.nCells:
                e = self.nCells
            store[s:e] = self.mat[:, s:e].todense().T
            s = e
        if e != self.nCells:
            raise AssertionError("ERROR: This is a bug in SparseToZarr. All cells might not have been successfully "
                                 "written into the zarr file. Please report this issue")


def subset_assay_zarr(zarr_fn: str, in_grp: str, out_grp: str,
                      cells_idx: np.ndarray, feat_idx: np.ndarray,
                      chunk_size: tuple):
    """
    Selects a subset of the data in an assay in the specified Zarr hierarchy.

    For the arguments `cells_idx` and `feat_idx`, refer to the documentation for numpy.split:
    https://numpy.org/doc/stable/reference/generated/numpy.split.html

    Args:
        zarr_fn: The file name for the Zarr hierarchy.
        in_grp: Group in Zarr hierarchy to subset.
        out_grp: Group name in Zarr hierarchy to write subsetted assay to.
        cells_idx: A list of cell indices to (keep | drop ?).
        feat_idx: A list of feature indices to (keep | drop ?).
        chunk_size: The requested size of chunks to load into memory and process.

    Returns:
        None
    """
    z = zarr.open(zarr_fn, 'r+')
    ig = z[in_grp]
    og = create_zarr_dataset(z, out_grp, chunk_size, 'uint32', (len(cells_idx), len(feat_idx)))
    pos_start, pos_end = 0, 0
    for i in tqdm(np.array_split(cells_idx, len(cells_idx) // chunk_size[0] + 1)):
        pos_end += len(i)
        og[pos_start:pos_end, :] = ig.get_orthogonal_selection((i, feat_idx))
        pos_start = pos_end
    return None


def dask_to_zarr(df, z, loc, chunk_size, nthreads: int, msg: str = None):
    # TODO: perhaps change name of Dask array so it does not get confused with a dataframe
    """
    Creates a Zarr hierarchy from a Dask array.

    Args:
        df (): Dask array.
        z (): Zarr hierarchy.
        loc (): Location to write data/Zarr hierarchy to.
        chunk_size (): Size of chunks to load into memory and process.
        nthreads (int): Number of threads to use.
        msg (str): Message to use with progress bar (Default: f"Writing data to {loc}").
    """
    if msg is None:
        msg = f"Writing data to {loc}"
    og = create_zarr_dataset(z, loc, chunk_size, 'float64', df.shape)
    pos_start, pos_end = 0, 0
    for i in tqdm(df.blocks, total=df.numblocks[0], desc=msg):
        pos_end += i.shape[0]
        og[pos_start:pos_end, :] = controlled_compute(i, nthreads)
        pos_start = pos_end
    return None


class SubsetZarr:

    def __init__(self, in_zarr: str, out_zarr: str, cell_key: str = None, cell_idx: np.ndarray = None,
                 reset_cell_filter: bool = True, overwrite_existing_file: bool = False,
                 overwrite_cell_data: bool = False) -> None:

        """
        Split Zarr file using a subset of cells

        Args:
            in_zarr: Path of input Zarr file to be subsetted.
            out_zarr: Path of output Zarr files containing only a subset of cells.
            cell_key: Name of a boolean column in cell metadata. The cells with with value True are included in the
                      subset.
            cell_idx: Indices of the cells to be included in the subsetted. Only used when cell_key is None.
            reset_cell_filter: If True, then the cell filtering information is removed, i.e. even the filtered out cells
                               are set as True as in the 'I' column. To keep the filtering information set the value for
                               this parameter to False. (Default value: True)
            overwrite_existing_file: If True, then overwrites the existing data. (Default value: False)
            overwrite_cell_data: If True, then overwrites cell data (Default value: True)
        Returns:
        """

        if cell_key is None and cell_idx is None:
            raise ValueError("Both 'cell_key' and 'cell_idx' parameters cannot be None")
        self.iZname = in_zarr
        self.oZname = out_zarr
        self.cellKey = cell_key
        self.cellIdx = cell_idx

        self.resetCells = reset_cell_filter
        self.overFn = overwrite_existing_file
        self.overcells = overwrite_cell_data

        self._check_files()
        self.iz = zarr.open(self.iZname)
        self._check_idx()
        self.oz = zarr.open(self.oZname, mode='w')
        self._prep_cell_data()
        self.assays = self._get_assays()
        self._prep_counts()

    def _check_files(self):
        if self.iZname == self.oZname:
            raise ValueError("You are trying to overwrite the current Zarr file itself with the subset. "
                             "This is not allowed. Please change the name/path of output file, by supplying a "
                             "different value to `out_zarr` parameter. No subsetting was performed")

        if os.path.isdir(self.oZname) and self.overFn is False:
            logger.error(f"Zarr file with name: {self.oZname} already exists.\nIf you want to overwrite it then please "
                         f"set overwrite_existing_file to True. No subsetting was performed.")
            return None

    def _check_idx(self):
        if self.cellIdx is None:
            idx = self.iz['cellData'][self.cellKey][:]
            if idx.dtype != bool:
                raise ValueError(f"ERROR: {self.cellKey} is not of boolean type. Cannot perform subsetting")
            self.cellIdx = np.where(idx)[0]

    def _prep_cell_data(self):
        n_cells = len(self.cellIdx)
        if 'cellData' in self.oz:
            g = self.oz['cellData']
        else:
            g = self.oz.create_group('cellData')

        for i in self.iz['cellData'].keys():
            if i in g and self.overcells is False:
                continue
            if i in ['I'] and self.resetCells:
                create_zarr_obj_array(g, 'I', [True for _ in range(n_cells)], 'bool')
                continue
            v = self.iz['cellData'][i][:][self.cellIdx]
            create_zarr_obj_array(g, i, v, dtype=v.dtype)

    def _get_assays(self):
        assays = []
        for i in self.iz.group_keys():
            if 'is_assay' in self.iz[i].attrs.keys():
                assays.append(i)
        return assays

    def _get_raw_data(self, assay_name):
        import dask.array as daskarr

        return daskarr.from_zarr(self.iz[assay_name]['counts'], inline_array=True)

    def _prep_counts(self):
        n_cells = len(self.cellIdx)
        for assay_name in self.assays:
            print (assay_name)
            raw_data = self._get_raw_data(assay_name)
            print (self.iz[assay_name]['featureData']['ids'][:])
            print (self.iz[assay_name]['featureData']['names'][:])
            create_zarr_count_assay(self.oz, assay_name, raw_data.chunksize, n_cells,
                                    self.iz[assay_name]['featureData']['ids'][:],
                                    self.iz[assay_name]['featureData']['names'][:], raw_data.dtype)

    def dump(self):
        for assay_name in self.assays:
            raw_data = self._get_raw_data(assay_name)
            store = self.oz[f"{assay_name}/counts"]
            s, e, = 0, 0
            for a in tqdm(raw_data[self.cellIdx].blocks,
                          desc=f"Subsetting assay: {assay_name}", total=raw_data.numblocks[0]):
                if a.shape[0] > 0:
                    e += a.shape[0]
                    store[s:e] = a.compute()
                    s = e


class ZarrMerge:
    """
    Merge multiple Zarr files into a single Zarr file.

    Attributes:
        assays: List of assay objects to be merged. For example, [ds1.RNA, ds2.RNA].
        names: Names of the each assay objects in the `assays` parameter.
        mergedCells:
        nCells: Number of cells in dataset.
        featCollection:
        mergedFeats:
        nFeats: Number of features in the dataset.
        featOrder:
        z: The merged Zarr file.
        assayGroup:
    """
    def __init__(self, zarr_path: str, assays: list, names: List[str], merge_assay_name: str,
                 chunk_size=(1000, 1000), dtype: str = None, overwrite: bool = False, prepend_text: str = 'orig',
                 reset_cell_filter: bool = True):
        """
        Args:
            zarr_path: Name of the new, merged Zarr file with path.
            assays: List of assay objects to be merged. For example, [ds1.RNA, ds2.RNA].
            names: Names of the each assay objects in the `assays` parameter. They should be in the same order as in
                   `assays` parameter.
            merge_assay_name: Name of assay in the merged Zarr file. For example, for scRNA-Seq it could be simply,
                              'RNA'.
            chunk_size: Tuple of cell and feature chunk size. (Default value: (1000, 1000)).
            dtype: Dtype of the raw values in the assay. Dtype is automatically inferred from the provided assays. If
                   assays have different dtypes then a float type is used.
            overwrite: If True, then overwrites previously created assay in the Zarr file. (Default value: False).
            prepend_text: This text is pre-appended to each column name (Default value: 'orig').
            reset_cell_filter: If True, then the cell filtering information is removed, i.e. even the filtered out cells
                               are set as True as in the 'I' column. To keep the filtering information set the value for
                               this parameter to False. (Default value: True)
        """
        self.assays = assays
        self.names = names
        self.mergedCells = self._merge_cell_table(reset_cell_filter, prepend_text)
        self.nCells = self.mergedCells.shape[0]
        self.featCollection = self._get_feat_ids(assays)
        self.mergedFeats = self._merge_order_feats()
        self.nFeats = self.mergedFeats.shape[0]
        self.featOrder = self._ref_order_feat_idx()
        self.z = self._use_existing_zarr(zarr_path, merge_assay_name, overwrite)
        self._ini_cell_data()
        if dtype is None:
            if len(set([str(x.rawData.dtype) for x in self.assays])) == 1:
                dtype = str(self.assays[0].rawData.dtype)
            else:
                dtype = 'float'
        self.assayGroup = create_zarr_count_assay(
            self.z['/'], merge_assay_name, chunk_size, self.nCells, list(self.mergedFeats.index),
            list(self.mergedFeats.names.values), dtype
        )

    def _merge_cell_table(self, reset, prepend_text):
        if len(self.assays) != len(set(self.names)):
            raise ValueError("ERROR: A unique name should be provided for each of the assay")
        if prepend_text == '':
            prepend_text = None
        ret_val = []
        for assay, name in zip(self.assays, self.names):
            a = assay.cells.to_pandas_dataframe(assay.cells.columns)
            a['ids'] = [f"{name}__{x}" for x in a['ids']]
            for i in a.columns:
                if i not in ['ids', 'I', 'names'] and prepend_text is not None:
                    a[f"{prepend_text}_{i}"] = assay.cells.fetch_all(i)
                    a = a.drop(columns=[i])
            if reset:
                a['I'] = np.ones(len(a['ids'])).astype(bool)
            ret_val.append(a)
        return pd.concat(ret_val).reset_index(drop=True)

    @staticmethod
    def _get_feat_ids(assays):
        ret_val = []
        for i in assays:
            ret_val.append(i.feats.to_pandas_dataframe(['names', 'ids']).set_index('ids')['names'].to_dict())
        return ret_val

    def _merge_order_feats(self):
        union_set = {}
        for ids in self.featCollection:
            for i in ids:
                if i not in union_set:
                    union_set[i] = ids[i]
        return pd.DataFrame({'idx': range(len(union_set)),
                             'names': list(union_set.values()),
                             'ids': list(union_set.keys())}).set_index('ids')

    def _ref_order_feat_idx(self):
        ret_val = []
        for ids in self.featCollection:
            ret_val.append(self.mergedFeats['idx'].reindex(ids).values)
        return ret_val

    def _use_existing_zarr(self, zarr_path, merge_assay_name, overwrite):
        try:
            z = zarr.open(zarr_path, mode='r')
            if 'cellData' not in z:
                raise ValueError(
                    f"ERROR: Zarr file with name {zarr_path} exists but seems corrupted. Either delete the "
                    "existing file or choose another path")
            if merge_assay_name in z:
                if overwrite is False:
                    raise ValueError(
                        f"ERROR: Zarr file `{zarr_path}` already contains {merge_assay_name} assay. Choose "
                        "a different zarr path or a different assay name. Otherwise set overwrite to True")
            try:
                if not all(z['cellData']['ids'][:] == np.array(self.mergedCells['ids'].values)):
                    raise ValueError(f"ERROR: order of cells does not match the one in existing file: {zarr_path}")
            except KeyError:
                raise ValueError(f"ERROR: 'cell data' in Zarr file {zarr_path} seems corrupted. Either delete the "
                                 "existing file or choose another path")
            return zarr.open(zarr_path, mode='r+')
        except ValueError:
            # So no zarr file with same name exists. Check if a non zarr folder with the same name exists
            if os.path.exists(zarr_path):
                raise ValueError(
                    f"ERROR: Directory/file with name `{zarr_path}`exists. Either delete it or use another name")
            # creating a new zarr file
            return zarr.open(zarr_path, mode='w')

    def _ini_cell_data(self):
        if 'cellData' not in self.z:
            g = self.z.create_group('cellData')
            for i in self.mergedCells:
                vals = self.mergedCells[i].values
                create_zarr_obj_array(g, i, vals, vals.dtype)
        else:
            logger.info(f"cellData already exists so skipping _ini_cell_data")

    def dump(self, nthreads=2):
        # TODO: add docstring
        pos_start, pos_end = 0, 0
        for assay, feat_order in zip(self.assays, self.featOrder):
            for i in tqdm(assay.rawData.blocks, total=assay.rawData.numblocks[0],
                          desc=f"Writing data to merged file"):
                pos_end += i.shape[0]
                a = np.ones((i.shape[0], self.nFeats))
                a[:, feat_order] = controlled_compute(i, nthreads)
                self.assayGroup[pos_start:pos_end, :] = a
                pos_start = pos_end


def to_h5ad(assay, h5ad_filename: str) -> None:
    """
    Save an assay as an h5ad file.

    Args:
        assay: Assay to save.
        h5ad_filename: Name for the h5ad file to be created.

    Returns:
        None
    """
    import h5py

    def save_attr(group, col, scarf_col, md):
        d = md.fetch_all(scarf_col)
        h5[group].create_dataset(col, data=d.astype(h5py.special_dtype(vlen=str)))

    h5 = h5py.File(h5ad_filename, 'w')
    for i in ['X', 'obs', 'var']:
        h5.create_group(i)

    n_feats_per_cell = assay.cells.fetch_all(f"{assay.name}_nFeatures").astype(int)
    tot_counts = int(n_feats_per_cell.sum())

    for i, s in zip(['indptr', 'indices', 'data'],
                    [assay.cells.N + 1, tot_counts, tot_counts]):
        h5['X'].create_dataset(i, (s,), chunks=True, compression='gzip', dtype=int)
    h5['X/indptr'][:] = np.array([0] + list(n_feats_per_cell.cumsum())).astype(int)
    s, e = 0, 0
    for i in tqdm(assay.rawData.blocks, total=assay.rawData.numblocks[0]):
        i = csr_matrix(i.compute()).astype(int)
        e += i.data.shape[0]
        h5['X/data'][s:e] = i.data
        h5['X/indices'][s:e] = i.indices
        s = e
    save_attr('obs', '_index', 'ids', assay.cells)
    save_attr('var', '_index', 'ids', assay.feats)
    save_attr('var', 'gene_short_name', 'names', assay.feats)

    attrs = {
        'encoding-type': 'csr_matrix',
        'encoding-version': '0.1.0',
        'shape': np.array([assay.cells.N, assay.feats.N])
    }
    for i, j in attrs.items():
        h5['X'].attrs[i] = j

    attrs = {
        '_index': '_index',
        'column-order': np.array(['_index'], dtype=object),
        'encoding-type': 'dataframe',
        'encoding-version': '0.1.0'
    }
    for i, j in attrs.items():
        h5['obs'].attrs[i] = j

    attrs = {
        '_index': '_index',
        'column-order': np.array(['_index', 'gene_short_name'], dtype=object),
        'encoding-type': 'dataframe',
        'encoding-version': '0.1.0'
    }
    for i, j in attrs.items():
        h5['var'].attrs[i] = j

    h5.close()
    return None


def to_mtx(assay, mtx_directory: str, compress: bool = False):
    """
    Save an assay as a Matrix Market directory.

    Args:
        assay: Scarf assay. For example: `ds.RNA`
        mtx_directory: Out directory where MTX file will be saved along with barcodes and features file
        compress: If True, then the files are compressed and saved with .gz extension. (Default value: False).

    Returns:
        None
    """
    from scipy.sparse import coo_matrix
    import gzip

    if os.path.isdir(mtx_directory) is False:
        os.mkdir(mtx_directory)

    n_feats_per_cell = assay.cells.fetch_all(f"{assay.name}_nFeatures").astype(int)
    tot_counts = int(n_feats_per_cell.sum())
    if compress:
        barcodes_fn = 'barcodes.tsv.gz'
        features_fn = 'features.tsv.gz'
        h = gzip.open(os.path.join(mtx_directory, 'matrix.mtx.gz'), 'wt')
    else:
        barcodes_fn = 'barcodes.tsv.gz'
        features_fn = 'genes.tsv'
        h = open(os.path.join(mtx_directory, 'matrix.mtx'), 'w')
    h.write("%%MatrixMarket matrix coordinate integer general\n% Generated by Scarf\n")
    h.write(f"{assay.feats.N} {assay.cells.N} {tot_counts}\n")
    s = 0
    for i in tqdm(assay.rawData.blocks, total=assay.rawData.numblocks[0]):
        i = coo_matrix((i.compute()))
        df = pd.DataFrame({'col': i.col + 1, 'row': i.row + s + 1, 'd': i.data})
        df.to_csv(h, sep=' ', header=False, index=False, mode='a')
        s += i.shape[0]
    h.close()
    assay.cells.to_pandas_dataframe(['ids']).to_csv(
        os.path.join(mtx_directory, barcodes_fn),
        sep='\t', header=False, index=False)

    assay.feats.to_pandas_dataframe(['ids', 'names']).to_csv(
        os.path.join(mtx_directory, features_fn),
        sep='\t', header=False, index=False)



"""
Utility functions for logging.
"""
from loguru import logger
import sys

logger.remove()
logger.add(sys.stdout, colorize=True, format="<level>{level}</level>: {message}", level="INFO")

__all__ = ['logger']



from tqdm import tqdm
import numpy as np
from threadpoolctl import threadpool_limits
from .utils import controlled_compute
from numpy.linalg import LinAlgError
from .logging_utils import logger

__all__ = ['AnnStream']


def fix_knn_query(indices: np.ndarray, distances: np.ndarray, ref_idx: np.ndarray):
    fixed_ind, fixed_dist = indices.copy()[:, 1:], distances.copy()[:, 1:]
    # Identify positions where first index is not a self loop
    mis_idx = indices[:, 0].reshape(1, -1)[0] != ref_idx
    n_mis = mis_idx.sum()
    if n_mis > 0:
        for n, i, j, k in zip(np.where(mis_idx)[0], ref_idx[mis_idx], indices[mis_idx], distances[mis_idx]):
            p = np.where(j == i)[0]
            if len(p) > 0:
                # p is the position of self loop. We exclude this position
                p = p[0]
                j = np.array(list(j[:p]) + list(j[p + 1:]))
                k = np.array(list(k[:p]) + list(k[p + 1:]))
            else:
                # No self found at all. Poor recall? simply remove the last k neighbour
                j = j[:-1]
                k = k[:-1]
            fixed_ind[n] = j
            fixed_dist[n] = k
    return fixed_ind, fixed_dist, n_mis


class AnnStream:
    def __init__(self, data, k: int, n_cluster: int, reduction_method: str,
                 dims: int, loadings: np.ndarray, use_for_pca: np.ndarray,
                 mu: np.ndarray, sigma: np.ndarray,
                 ann_metric: str, ann_efc: int, ann_ef: int, ann_m: int,
                 nthreads: int, ann_parallel: bool, rand_state: int,
                 do_kmeans_fit: bool, disable_scaling: bool, ann_idx):
        self.data = data
        self.k = k
        if self.k >= self.data.shape[0]:
            self.k = self.data.shape[0] - 1
        self.nClusters = max(n_cluster, 2)
        self.dims = dims
        self.loadings = loadings
        if self.dims is None and self.loadings is None:
            raise ValueError("ERROR: Provide either value for atleast one: 'dims' or 'loadings'")
        self.annMetric = ann_metric
        self.annEfc = ann_efc
        self.annEf = ann_ef
        self.annM = ann_m
        self.nthreads = nthreads
        if ann_parallel:
            self.annThreads = self.nthreads
        else:
            self.annThreads = 1
        self.randState = rand_state
        self.batchSize = self._handle_batch_size()
        self.method = reduction_method
        self.nCells, self.nFeats = self.data.shape
        self.clusterLabels: np.ndarray = np.repeat(-1, self.nCells)
        disable_reduction = False
        if self.dims < 1:
            disable_reduction = True
        with threadpool_limits(limits=self.nthreads):
            if self.method == 'pca':
                self.mu, self.sigma = mu, sigma
                if self.loadings is None or len(self.loadings) == 0:
                    if len(use_for_pca) != self.nCells:
                        raise ValueError("ERROR: `use_for_pca` does not have sample length as nCells")
                    if disable_reduction is False:
                        self._fit_pca(disable_scaling, use_for_pca)
                else:
                    # Even though the dims might have been already adjusted according to loadings before calling
                    # AnnStream, it could still be overwritten by _handle_batch_size. Hence need to hard set it here.
                    self.dims = self.loadings.shape[1]
                    # it is okay for dimensions to be larger than batch size here because we will not fit the PCA
                if disable_scaling:
                    if disable_reduction:
                        self.reducer = lambda x: x
                    else:
                        self.reducer = lambda x: x.dot(self.loadings)
                else:
                    if disable_reduction:
                        self.reducer = lambda x: self.transform_z(x)
                    else:
                        self.reducer = lambda x: self.transform_z(x).dot(self.loadings)
            elif self.method == 'lsi':
                if self.loadings is None or len(self.loadings) == 0:
                    if disable_reduction is False:
                        self._fit_lsi()
                else:
                    self.dims = self.loadings.shape[1]
                if disable_reduction:
                    self.reducer = lambda x: x
                else:
                    self.reducer = lambda x: x.dot(self.loadings)
            elif self.method == 'custom':
                if self.loadings is None or len(self.loadings) == 0:
                    logger.warning("No loadings provided for manual dimension reduction")
                else:
                    self.dims = self.loadings.shape[1]
                if disable_reduction:
                    self.reducer = lambda x: x
                else:
                    self.reducer = lambda x: x.dot(self.loadings)
            else:
                raise ValueError(f"ERROR: Unknown reduction method: {self.method}")
            if ann_idx is None:
                self.annIdx = self._fit_ann()
            else:
                self.annIdx = ann_idx
                self.annIdx.set_ef(self.annEf)
                self.annIdx.set_num_threads(1)
            self.kmeans = self._fit_kmeans(do_kmeans_fit)

    def _handle_batch_size(self):
        if self.dims > self.data.shape[0]:
            self.dims = self.data.shape[0]
        batch_size = self.data.chunksize[0]  # Assuming all chunks are same size
        if self.dims >= batch_size:
            self.dims = batch_size - 1  # -1 because we will do PCA +1
            logger.info(f"Number of PCA/LSI components reduced to batch size of {batch_size}")
        if self.nClusters > batch_size:
            self.nClusters = batch_size
            logger.info(f"Cluster number reduced to batch size of {batch_size}")
        return batch_size

    def iter_blocks(self, msg: str = '') -> np.ndarray:
        for i in tqdm(self.data.blocks, desc=msg, total=self.data.numblocks[0]):
            yield controlled_compute(i, self.nthreads)

    def transform_z(self, a: np.ndarray) -> np.ndarray:
        return (a - self.mu) / self.sigma

    def transform_ann(self, a: np.ndarray, k: int = None, self_indices: np.ndarray = None) -> tuple:
        if k is None:
            k = self.k
        # Adding +1 to k because first neighbour will be the query itself
        if self_indices is None:
            i, d = self.annIdx.knn_query(a, k=k)
            return i, d
        else:
            i, d = self.annIdx.knn_query(a, k=k + 1)
            return fix_knn_query(i, d, self_indices)

    def _fit_pca(self, disable_scaling, use_for_pca) -> None:
        from sklearn.decomposition import IncrementalPCA
        # We fit 1 extra PC dim than specified and then ignore the last PC.
        self._pca = IncrementalPCA(n_components=self.dims + 1, batch_size=self.batchSize)
        do_sample_subset = False if use_for_pca.sum() == self.nCells else True
        s, e = 0, 0
        # We store the first block of values here. if such a case arises that we are left with less dims+1 cells to fit
        # then those cells can be added to end_reservoir for fitting. if there are no such cells then end reservoir is
        # just by itself after fitting rest of the cells. If may be the case that the first batch itself has less than
        # dims+1 cells. in that we keep adding cells to carry_over pile until it is big enough.
        end_reservoir = []
        # carry_over store cells that can yet not be added to end_reservoir ot be used for fitting pca directly.
        carry_over = []
        for i in self.iter_blocks(msg='Fitting PCA'):
            if do_sample_subset:
                e = s + i.shape[0]
                i = i[use_for_pca[s:e]]
                s = e
            if disable_scaling is False:
                i = self.transform_z(i)
            if len(carry_over) > 0:
                i = np.vstack((carry_over, i))
                carry_over = []
            if len(i) < (self.dims + 1):
                carry_over = i
                continue
            if len(end_reservoir) == 0:
                end_reservoir = i
                continue
            try:
                self._pca.partial_fit(i, check_input=False)
            except LinAlgError:
                # Add retry counter to make memory consumption doesn't escalate
                carry_over = i
        if len(carry_over) > 0:
            i = np.vstack((end_reservoir, carry_over))
        else:
            i = end_reservoir
        try:
            self._pca.partial_fit(i, check_input=False)
        except LinAlgError:
            logger.warning("{i.shape[0]} samples were not used in PCA fitting due to LinAlgError", flush=True)
        self.loadings = self._pca.components_[:-1, :].T

    def _fit_lsi(self) -> None:
        from gensim.models import LsiModel
        from gensim.matutils import Dense2Corpus

        self._lsiModel = LsiModel(Dense2Corpus(controlled_compute(self.data.blocks[0], self.nthreads).T),
                                  num_topics=self.dims, chunksize=self.data.chunksize[0],
                                  id2word={x: x for x in range(self.data.shape[1])}, extra_samples=0)
        for n, i in enumerate(self.iter_blocks(msg="Fitting LSI model")):
            if n == 0:
                continue
            self._lsiModel.add_documents(Dense2Corpus(i.T))
        self.loadings = self._lsiModel.get_topics().T

    def _fit_ann(self):
        import hnswlib

        dims = self.dims
        if dims < 1:
            dims = self.data.shape[1]
        ann_idx = hnswlib.Index(space=self.annMetric, dim=dims)
        ann_idx.init_index(max_elements=self.nCells, ef_construction=self.annEfc,
                           M=self.annM, random_seed=self.randState)
        ann_idx.set_ef(self.annEf)
        ann_idx.set_num_threads(self.annThreads)
        for i in self.iter_blocks(msg='Fitting ANN'):
            ann_idx.add_items(self.reducer(i))
        return ann_idx

    def _fit_kmeans(self, do_ann_fit):
        from sklearn.cluster import MiniBatchKMeans

        if do_ann_fit is False:
            return None
        kmeans = MiniBatchKMeans(
            n_clusters=self.nClusters, random_state=self.randState,
            batch_size=self.batchSize)
        with threadpool_limits(limits=self.nthreads):
            for i in self.iter_blocks(msg='Fitting kmeans'):
                kmeans.partial_fit(self.reducer(i))
        temp = []
        for i in self.iter_blocks(msg='Estimating seed partitions'):
            temp.extend(kmeans.predict(self.reducer(i)))
        self.clusterLabels = np.array(temp)
        return kmeans



"""
- Methods:
    - meld_assay:
    - make_bed_from_gff: create pybedtools object for genes from a GFF file
"""
# TODO: add description in docstring

from tqdm import tqdm
from typing import List, Dict
import pandas as pd
import numpy as np
from .writers import create_zarr_count_assay
from .utils import controlled_compute
from .logging_utils import logger

__all__ = ['meld_assay', 'make_bed_from_gff']


def make_bed_from_gff(gff: str, up_offset: int = 2000,
                      valid_ids: List[str] = None, flavour: str = 'body'):
    """Create pybedtools object for genes from a GFF file. Gene coordinates are promoter extended."""
    try:
        from pybedtools import BedTool
    except ImportError:
        raise ImportError("pybedtools is not installed. Check out this link to install"
                          " https://daler.github.io/pybedtools/main.html#install-via-conda")
    out = []
    ignored_genes = 0
    unknown_ids = 0
    if valid_ids is not None:
        valid_ids = {x: None for x in valid_ids}
    with open(gff) as h:
        # Testing whether first 5 lines are comment lines
        for i in range(5):
            l = next(h)
            if l[0] != '#':
                logger.warning(f"line num {i} is not comment line", flush=True)
        for l in tqdm(h):
            c = l.split('\t')
            if c[2] != 'gene':
                continue
            a = [x.split(' ') for x in c[8].rstrip('\n').split('; ')]
            a = {x: y.strip('"') for x, y in a}
            if 'gene_id' not in a:
                unknown_ids += 1
                continue
            if valid_ids is not None and a['gene_id'] not in valid_ids:
                ignored_genes += 1
                continue
            # Fetch start and end coordinate
            s, e = int(c[3]), int(c[4])
            if flavour == 'body':
                if c[6] == '+':
                    s = s - up_offset
                    s = max(s, 0)
                elif c[6] == '-':
                    e = e + up_offset
                else:
                    raise ValueError('ERROR: Unsupported symbol for strand')
            elif flavour == 'promoter':
                if c[6] == '+':
                    e = s + up_offset
                    s = s - up_offset
                    s = max(s, 0)
                elif c[6] == '-':
                    s = e - up_offset
                    s = max(s, 0)
                    e = e + up_offset
                else:
                    raise ValueError('ERROR: Unsupported symbol for strand')
            else:
                raise ValueError('ERROR: `flavour` can either be `body` or `promoter`')
            if c[0].startswith('chr'):
                chrom = c[0]
            else:
                chrom = f'chr{c[0]}'
            if 'gene_name' in a:
                gn = a['gene_name']
            else:
                gn = a['gene_id']
            o = '\t'.join([chrom, str(s), str(e), a['gene_id'], gn, c[6]])
            out.append(o)
    logger.info(f"{len(out)} genes found in the GFF file")
    logger.info(f"{ignored_genes} genes were ignored as they were not present in the valid_ids")
    logger.info(f"{unknown_ids} genes were ignored as they did not have gene_id column")
    return BedTool('\n'.join(out), from_string=True)


def _create_bed_from_coord_ids(ids: List[str], invalid_ids: Dict[str, None]):
    """convert list of 'chr:start-end' format strings to pybedtools object"""
    try:
        # noinspection PyUnresolvedReferences
        from pybedtools import BedTool
    except ImportError:
        raise ImportError("ERROR: pybedtools is not installed")

    out = []
    for i in ids:
        if i in invalid_ids:
            continue
        j = i.split(':')
        o = [j[0], j[1].split('-')[0], j[1].split('-')[1], i]
        out.append('\t'.join(o))
    return BedTool('\n'.join(out), from_string=True)


def _get_merging_map(a, b, b_name_pos: int = 7) -> dict:
    """Intersect BED files to obtain a dict with b names as keys and
       overlapped a names for each b in values as list
    """
    joined = a.intersect(b, wao=True)
    res = {x.name: [] for x in b}
    for i in joined:
        if i[b_name_pos] != '.':
            res[i[b_name_pos]].append(i.name)
    return res


def _convert_ids_to_idx(ids: pd.Series, cross_id_map: dict) -> dict:
    """
    Convert ids to indices from scarf feats table.
    """
    ref = {v: k for k, v in ids.to_dict().items()}
    idx = {x: [] for x in cross_id_map}
    null_ids = 0
    for i in cross_id_map:
        if len(cross_id_map[i]) > 0:
            idx[i] = [ref[x] for x in cross_id_map[i]]
        else:
            null_ids += 1
    logger.info(f"{null_ids}/{len(cross_id_map)} ids did not have a cross id")
    return idx


def _create_counts_mat(assay, out_store, feat_order: list, cross_idx_map: dict, nthreads: int) -> None:
    c_pos_start = 0
    for a in tqdm(assay.rawData.blocks, total=assay.rawData.numblocks[0]):
        a = controlled_compute(a, nthreads)
        b = np.zeros((a.shape[0], len(feat_order)))
        c_pos_end = c_pos_start + a.shape[0]
        for n, i in enumerate(feat_order):
            if i in cross_idx_map and len(cross_idx_map[i]) > 0:
                b[:, n] = a[:, cross_idx_map[i]].sum(axis=1)
        out_store[c_pos_start:c_pos_end, :] = b
        c_pos_start = c_pos_end
    return None


def meld_assay(assay, reference_bed, out_name: str, nthreads: int,
               peaks_col: str = 'ids', ignore_ids: List[str] = None):
    # TODO: add docstring
    if ignore_ids is None:
        ignore_ids = {}
    else:
        ignore_ids = {x: None for x in ignore_ids}
    peaks_bed = _create_bed_from_coord_ids(assay.feats.fetch_all(peaks_col), invalid_ids=ignore_ids)
    cross_idx_map = _convert_ids_to_idx(pd.Series(assay.feats.fetch_all(peaks_col)),
                                        _get_merging_map(peaks_bed, reference_bed))
    feat_ids = [x[3] for x in reference_bed]
    feat_names = [x[4] for x in reference_bed]
    g = create_zarr_count_assay(z=assay.z['/'], assay_name=out_name, chunk_size=assay.rawData.chunksize,
                                n_cells=assay.rawData.shape[0], feat_ids=feat_ids, feat_names=feat_names)
    _create_counts_mat(assay=assay, out_store=g, feat_order=feat_ids, cross_idx_map=cross_idx_map, nthreads=nthreads)



from setuptools import setup, find_packages
from setuptools.command.install import install
import os
import glob


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


class PostInstallCommand(install):
    """
    Post-installation for installation mode.
    From here:
    https://github.com/benfred/py-spy/blob/290584dde76834599d66d74b64165dfe9a357ef5/setup.py#L42
    """
    def run(self):
        install.run(self)
        if not os.path.isdir(self.install_scripts):
            os.makedirs(self.install_scripts)
        source_dir = os.path.dirname(os.path.abspath(__file__))
        source_files = glob.glob(os.path.join(source_dir, "bin")+'/*')
        for source in source_files:
            target = os.path.join(self.install_scripts, os.path.basename(source))
            if os.path.isfile(target):
                os.remove(target)
            with open(source, 'rb') as src, open(target, 'wb') as dst:
                dst.write(src.read())
            os.chmod(target, 0o655)


if __name__ == "__main__":
    classifiers = [
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
    ]
    keywords = ['store']
    version = open('VERSION').readline().rstrip('\n')
    install_requires = ['pybind11'] + [x.strip() for x in open('requirements.txt')] + \
                       ['dask[array]', 'dask[dataframe]']
    dependency_links = ['https://github.com/fraenkel-lab/pcst_fast/tarball/master#egg=pcst_fast-1.0.7']
    setup(
        name='scarf',
        description='Scarf',
        long_description=read('README.rst'),
        author='Parashar Dhapola',
        author_email='parashar.dhapola@gmail.com',
        url='https://github.com/parashardhapola/scarf',
        license='BSD 3-Clause',
        classifiers=classifiers,
        keywords=keywords,
        install_requires=install_requires,
        dependency_links=dependency_links,
        version=version,
        packages=find_packages(),
        include_package_data=False,
        cmdclass={'install': PostInstallCommand}
    )



# Author: Leland McInnes <leland.mcinnes@gmail.com>
# Modified file
#
# License: BSD 3 clause

import locale
locale.setlocale(locale.LC_NUMERIC, "C")


__all__ = ['fit', 'transform', 'fit_transform']


def simplicial_set_embedding(
        g, embedding, n_epochs, a, b, random_seed, gamma,
        initial_alpha, negative_sample_rate, parallel, nthreads):
    import numba
    from threadpoolctl import threadpool_limits
    from umap.umap_ import make_epochs_per_sample
    from umap.layouts import optimize_layout_euclidean
    from sklearn.utils import check_random_state
    import numpy as np

    g.data[g.data < (g.data.max() / float(n_epochs))] = 0.0
    g.eliminate_zeros()
    epochs_per_sample = make_epochs_per_sample(g.data, n_epochs)
    head = g.row
    tail = g.col
    rng_state = check_random_state(random_seed).randint(
        np.iinfo(np.int32).min + 1, np.iinfo(np.int32).max - 1, 3).astype(np.int64)
    # Since threadpool_limits doesnt work well with numba. We will use numba's set_num_threads to limit threads
    if numba.config.NUMBA_NUM_THREADS > nthreads:
        numba.set_num_threads(nthreads)
    with threadpool_limits(limits=nthreads):
        embedding = optimize_layout_euclidean(
            embedding, embedding, head, tail, n_epochs, g.shape[1],
            epochs_per_sample, a, b, rng_state, gamma, initial_alpha,
            negative_sample_rate, parallel=parallel, verbose=True)
    return embedding


def fuzzy_simplicial_set(g, set_op_mix_ratio):
    tg = g.transpose()
    prod = g.multiply(tg)
    res = (
        set_op_mix_ratio * (g + tg - prod)
        + (1.0 - set_op_mix_ratio) * prod
    )
    res.eliminate_zeros()
    return res.tocoo()


def fit(graph, embedding, spread, min_dist,
        set_op_mix_ratio, n_epochs, random_seed,
        repulsion_strength, initial_alpha, negative_sample_rate, parallel, nthreads):
    from umap.umap_ import find_ab_params

    a, b = find_ab_params(spread, min_dist)
    sym_graph = fuzzy_simplicial_set(graph, set_op_mix_ratio)
    embedding = simplicial_set_embedding(
        sym_graph, embedding, n_epochs, a, b, random_seed, repulsion_strength,
        initial_alpha, negative_sample_rate, parallel, nthreads)
    return embedding, a, b


def transform(graph, embedding,
              a, b, n_epochs, random_seed,
              repulsion_strength, initial_alpha, negative_sample_rate, parallel, nthreads):
    return simplicial_set_embedding(
        graph, embedding, n_epochs, a, b, random_seed,
        repulsion_strength, initial_alpha, negative_sample_rate, parallel, nthreads)


def fit_transform(graph, ini_embed, spread: float, min_dist: float, tx_n_epochs: int, fit_n_epochs: int,
                  random_seed: int, set_op_mix_ratio: float = 1.0, repulsion_strength: float = 1.0,
                  initial_alpha: float = 1.0, negative_sample_rate: float = 5, parallel: bool = False,
                  nthreads: int = 1):
    e, a, b = fit(graph, ini_embed,
                  spread=spread, min_dist=min_dist, set_op_mix_ratio=set_op_mix_ratio,
                  n_epochs=fit_n_epochs, random_seed=random_seed, repulsion_strength=repulsion_strength,
                  initial_alpha=initial_alpha, negative_sample_rate=negative_sample_rate,
                  parallel=parallel, nthreads=nthreads)
    t = transform(graph, e, a, b, n_epochs=tx_n_epochs, random_seed=random_seed,
                  repulsion_strength=repulsion_strength, initial_alpha=initial_alpha,
                  negative_sample_rate=negative_sample_rate, parallel=parallel, nthreads=nthreads)
    return t
